#!/usr/bin/perl -w

# use for dump of user's mysql databases
use vars qw($mysqlUserUser $mysqlUserServer $mysqlUserPw);

use vars qw( %bin
             $ptrArgs
             $dumpDir
             $workDir
          $storage);

# Prefer internal modules over external
use lib '.';

use strict;

use PleskVersion;
use PleskMaxVersion;
use PleskStructure;
use Status;
use SiteApp;
use Mailman;
use AgentConfig;

require "agent.include.pl";

#
# parameters for connection to database
#
my $dbUser = 'admin';
my $dbName = 'psa';
my $dbHost = 'localhost';
my $dbType = 'mysql';
#
# end parameters for connection to database
#

#
# maps to translate xmlName => fieldName
#

my %xmlPermissions = ('create_domains'=>1,
		      'create_domains'=>1,
		      'manage_phosting'=>1,
		      'manage_quota'=>1,
		      'manage_subdomains'=>1,
		      'change_limits'=>1,
		      'manage_dns'=>1,
		      'manage_log'=>1,
		      'manage_crontab'=>1,
		      'manage_anonftp'=>1,
		      'manage_webapps'=>1,
		      'manage_sh_access'=>1,
		      'manage_maillists'=>1,
		      'make_dumps'=>1,
		      'manage_drweb'=>1,
		      'make_phosting'=>1,
		      'manage_quota'=>1,
		      'manage_not_chroot_shell'=>1,
		      'manage_ftp_password'=>1,
		      'cp_access'=>1,
		      'remote_access_interface'=>1,
		      'manage_domain_aliases'=>1,
                      'manage_dashboard'=>1,
                      'dashboard'=>1,
                      'stdgui'=>1
		     );

my %commonLimits = ('max_wu' => 1,
					'max_db' => 1,
					'max_pbox' => 1,
					'mbox_quota' => 1,
					'max_redir' => 1,
					'max_mg' => 1,
					'max_resp' => 1,
					'disk_space' => 1,
					'max_traffic' => 1
				   );

my %domainLimits25 = %commonLimits;
$domainLimits25{'max_box'} = 1;

my %domainLimits = %commonLimits;
$domainLimits{'max_box'} = 1;
$domainLimits{'expiration'} = 1;
$domainLimits{'max_subdom'} = 1;
$domainLimits{'max_maillists'} = 1;
$domainLimits{'max_webapps'} = 1;

my %clientLimits25 = %domainLimits25;
$clientLimits25{'max_dom'} = 1;

my %clientLimits = %domainLimits;
$clientLimits{'max_dom'} = 1;

my %typeOfField = ('expiration'=>'timestamp');

my %clientsAttribute = ('name'=>'login',
						'contact'=>'pname',
					   );

my %clientsInfo = ('company'=>'cname',
				   'phone'=>'phone',
				   'fax'=>'fax',
				   'address'=>'address',
				   'city'=>'city',
				   'state'=>'state',
				   'zip'=>'pcode',
				   'country'=>'country',
				   'email'=>'email',
				  );

my %domainUserAttribute = ('contact'=>'personalName',);

my %domainUserInfo = ('company'=>'companyName',
					  'phone'=>'phone',
					  'fax'=>'fax',
					  'address'=>'address',
					  'city'=>'city',
					  'state'=>'state',
					  'zip'=>'zip',
					  'email'=>'email',
					  'country'=>'country',);

my %hostingAttribute = ('https'=>'ssl',
						'fp'=>'fp',
						'fpssl'=>'fp_ssl',
						'fpauth'=>'fp_enable',
						'webstat'=>'webstat',
						'at-domains'=>'at_domains',
					   );

my %hostingScripting = ('ssi'=>'ssi',
						'php'=>'php',
						'php_safe_mode'=>'php_safe_mode',
						'cgi'=>'cgi',
						'perl'=>'perl',
						'asp'=>'asp',
						'python'=>'python',
						'coldfusion'=>'coldfusion',
						'asp_dot_net'=>'asp_dot_net',
					   );

my %webUserScripting = ('ssi'=>'ssi',
						'php'=>'php',
						'cgi'=>'cgi',
						'perl'=>'perl',
						'asp'=>'asp',
						'python'=>'python',
						'asp_dot_net'=>'asp_dot_net',
					   );

my %subDomainScripting = ('ssi'=>'ssi',
						  'php'=>'php',
						  'cgi'=>'cgi',
						  'perl'=>'perl',
						  'asp'=>'asp',
						  'python'=>'python',
						  'coldfusion'=>'coldfusion',
						  'asp_dot_net'=>'asp_dot_net',
						 );

my %emptyableBoolElementInTemplate = (
									  'asp' => 1,
									  'cgi' => 1,
									  'coldfusion' => 1,
									  'pdir_plesk_stat' => 1,
									  'perl' => 1,
									  'php' => 1,
									  'python' => 1,
									  'ssi' => 1
									 );

my %custom_button_owner_types = (
								 "client" => 4,       #IS_CLIENT
								 "domain-admin" => 8, #IS_DOMAIN_OWNER
								 "mailuser" => 16,    #IS_MAIL_USER
								);

my %locale_map = ('bg' => 'bg', 'hk' => 'zh-HK',
				'ru' => 'ru-RU', 'en' => 'en-US',
				'nl' => 'nl-NL', 'br' => 'pt-BR',
				'it' => 'it-IT', 'tr' => 'tr-TR',
				'es' => 'es-ES', 'pl' => 'pl-PL',
				'ca' => 'ca-ES', 'jp' => 'ja-JP',
				'tw' => 'zh-TW', 'fi' => 'fi-FI',
				'cn' => 'zh-CN', 'ko' => 'ko-KR',
				'fr' => 'fr-FR', 'pt' => 'pt-PT',
				'de' => 'de-DE', 'lt' => 'lt-LT'
				);

# list of attributes for template-item node in dump.xml
# needs to synchronize with template-item's ATTLIST of plesk.dtd
my %template_items = ('webmail' => 1, 'disk_space' => 1,
					'stat_ttl' => 1, 'maillists' => 1,
					'nonexist_mail' => 1, 'catch_addr' => 1,
					'bounce_mess' => 1, 'max_traffic' => 1,
					'max_box' => 1, 'mbox_quota' => 1,
					'max_redir' => 1, 'max_mg' => 1,
					'max_resp' => 1, 'max_wu' => 1,
					'max_db' => 1, 'max_maillists' => 1,
					'max_webapps' => 1, 'max_subdom' => 1,
					'expiration' => 1, 'vh_type' => 1,
					'quota' => 1, 'ssl' => 1,
					'same_ssl' => 1, 'fp' => 1,
					'fp_ssl' => 1, 'fp_auth' => 1,
					'asp' => 1, 'ssi' => 1,
					'php' => 1, 'cgi' => 1,
					'perl' => 1, 'python' => 1,
					'coldfusion' => 1, 'webstat' => 1,
					'errdocs' => 1, 'shell' => 1,
					'pdir_plesk_stat' => 1, 'dns_type' => 1
				   );

#
# end maps
#

#
# GLOBAL
#

#-----
# Array of custom buttons IDs, which are dumped within site applications
#-----
my @skip_custom_buttons;

#
# parse command line
#
my %arg_opts = ('--help|-h'=>'',
                '--get-status|-s'=>'',
                '--dump-accounts|-dc'=>'s',
                '--dump-domains|-dd'=>'s',
                '--dump-all|-da'=>'',
                '--get-content-list|-lc'=>'',
                '--no-content|-nc'=>'',
                '--only-mail|-om'=>'',
                '--no-compress|-nz'=>'',
                '--configuration-only|-co'=>'',
                '--server'=>'',
                '--output|-o'=>'s',
                '--description|-d'=>'s'
               );

%bin = ('tar'=>'tar','bsdtar'=>'bsdtar','split'=>'split','cp'=>'cp','rm'=>'rm','echo'=>'echo',
        'mysql'=>'mysql','mysqldump'=>'mysqldump',
        'psql'=>'psql','pg_dump'=>'pg_dump','uptime'=>'uptime','uname'=>'uname',
        'cat'=>'cat','pwd'=>'pwd','md5sum'=>'md5sum',
        'grep'=>'grep','find'=>'find','gzip'=>'gzip','wc'=>'wc',
        'locale'=>'locale', 'iconv'=>'iconv', 'python'=>'python');

$ptrArgs = getArguments(\@ARGV,\%arg_opts);

#
# search binaries
#

&getBins(\%bin,'/usr/local/psa/python2.2/bin:/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin:'.$ENV{'PATH'});

$workDir = `$bin{'pwd'}`;
chomp $workDir;

my $dumpFile = $workDir.'/dump.xml';
my $statusFile = $workDir.'/dumping-status.xml';
$dumpDir = $workDir.'/pma';

my $objDumpStatus = makeDumpStatus($statusFile);

#
# get MIME Base64 encoder
#
my $wrapBase64 = makeMIMEBase64();

my %domainsInfo;

AgentConfig::init() or die "Error: Plesk config file 'psa.conf' is not found\n";
AgentConfig::tarBin();
AgentConfig::mysqlBin();
PleskVersion::init(AgentConfig::get('PRODUCT_ROOT_D'));

#FIXME: Why only printAgentStatus checks for Plesk version?
if (exists $ptrArgs->{'get-status'}) {
  printAgentStatus();
  exit 0;
}

#
# get db connect
#
# It's better to always parse psa.conf to prevent troubles with Standard builds
my $mysqlpath = AgentConfig::get("MYSQL_BIN_D");
if ($mysqlpath) {
	$bin{'mysql'}     = $mysqlpath . "/mysql";
	$bin{'mysqldump'} = $mysqlpath . "/mysqldump";
}

$bin{'psql'} = AgentConfig::psqlBin() || 'psql';

my $dbAddress = "DBI:$dbType:dbname=$dbName;host=$dbHost";
my $wrapDbh = &getDbConnect(\%bin,$dbType,$dbAddress,
			   $dbUser, AgentConfig::get('password'),
			   $dbName, $dbHost, undef, undef, undef, PleskVersion::atLeast(7, 1, 0));

unless(ref($wrapDbh)=~/HASH/ && ref($wrapDbh->{'EXECUTE'})=~/CODE/){
  die "Error: can not connect to Plesk database\n";
}

PleskStructure::init($wrapDbh);

# Returns the elements that present in first array, but not in the second.
# Arrays must be sorted.
# Linear complexity.

sub arrayDifference {
  my ($a1, $a2) = @_;
  my @ret;

  my $i1 = 0; my $i2 = 0;
  while ($i1 < scalar(@$a1) && $i2 < scalar(@$a2)) {
    if ($a1->[$i1] eq $a2->[$i2]) { $i1++; $i2++; next; }
    if ($a1->[$i1] lt $a2->[$i2]) { push @ret, $a1->[$i1]; $i1++; next; }
    if ($a1->[$i1] gt $a2->[$i2]) { $i2++; next; }
  }

  while ($i1 < scalar(@$a1)) {
    push @ret, $a1->[$i1];
    $i1++;
  }
  return @ret;
}

# Returns the elements that present in both arrays
# Arrays must be storted
# Linear complexity

sub arrayIntersection {
  my ($a1, $a2) = @_;
  my @ret;

  my $i1 = 0; my $i2 = 0;
  while ($i1 < scalar(@$a1) && $i2 < scalar(@$a2)) {
    if ($a1->[$i1] eq $a2->[$i2]) { push @ret, $a1->[$i1]; $i1++; $i2++; next; }
    if ($a1->[$i1] lt $a2->[$i2]) { $i1++; next; }
    if ($a1->[$i1] gt $a2->[$i2]) { $i2++; next; }
  }

  return @ret;
}


if (exists $ptrArgs->{'get-content-list'}) {
  makeContentList();
  exit 0;
}

if (exists ($ptrArgs->{'dump-all'}) ||
    exists ($ptrArgs->{'dump-accounts'}) ||
    exists ($ptrArgs->{'dump-domains'})) {

  initStorage();

  getDumpDir($dumpDir);

  if (exists $ptrArgs->{'no-compress'}) {
	setCompress(0);
  }

  printToLog("Version of Plesk: " . PleskVersion::get());

  if (PleskVersion::atLeast(7, 1, 0)) {
	setCharset("UTF-8");
  }

  printToLog("Work dir: $workDir");
  printToLog("Dump file: $dumpFile");
  printToLog("Status file: ".$objDumpStatus->{'FILE'}->());

  my (@clients, @domains);

  if (exists $ptrArgs->{'dump-accounts'}) {
    my $value = $ptrArgs->{'dump-accounts'};
  	if ($value eq "-") {
		$value = <STDIN>;
		chomp $value;
	}
    my @inputClients = sort { $a cmp $b } split(/\s*,\s*/, $value);
    my @pleskClients = sort { $a cmp $b } PleskStructure::getClients();
    my @missingClients = arrayDifference(\@inputClients, \@pleskClients);

    if (@missingClients) {
      die "The following clients were not found on the host: " . (join ",", @missingClients);
    }

    @clients = @inputClients;
  }

  if (exists $ptrArgs->{'dump-domains'}) {
    my $value = $ptrArgs->{'dump-domains'};
  	if ($value eq "-") {
		$value = <STDIN>;
		chomp $value;
	}
    my @inputDomains = sort { $a cmp $b } split(/\s*,\s*/,$value);
    my @pleskDomains = sort { $a cmp $b } PleskStructure::getDomains();
    my @missingDomains = arrayDifference(\@inputDomains, \@pleskDomains);

    if (@missingDomains) {
      Logging::warning("The following domains were not found on the host: " . (join ",", @missingDomains));
    }

    @domains = @inputDomains;
  }

  my $root = XmlNode('migration-dump', 'attributes' => {'agent-name' => 'PleskX'});
  if (exists $ptrArgs->{'description'}) {
    $root->{'ADDCHILD'}->(XmlNode('description', 'content' => $ptrArgs->{'description'}));
  }

  if (exists $ptrArgs->{'dump-all'}) {
    createFullDump($root);
  } elsif (exists $ptrArgs->{'dump-accounts'}) {
    my %clientsDesc;

    # if domains & accounts specified simultaneously, then it's a backup and only selected
    # domains should be included in backup
    if (!exists $ptrArgs->{'dump-domains'}) {
      @domains = sort { $a cmp $b } PleskStructure::getDomains();
    }

    my $client;
    foreach $client (@clients) {
      my @clientDomains = PleskStructure::getDomainsForClient($client);
      @clientDomains = sort { $a cmp $b } @clientDomains;

      my @domainsToDump = arrayIntersection(\@clientDomains, \@domains);
      $clientsDesc{$client} = \@domainsToDump;
    }

    createClientsDump($root, \%clientsDesc);

  } elsif (exists $ptrArgs->{'dump-domains'}) {
    createDomainsDump($root, \@domains);
  } else {
    die "No action supplied";
  }

  if (exists $ptrArgs->{'server'}) {
    makeServerNode($root);
  }

  exit($storage->finish($root));
} else {
  printHelp();
}

exit 0;

#
# end main
#
#==============================================================
#
# subroutines
#

sub createFullDump {
  my ($root) = @_;

  my @clients = PleskStructure::getClients();

  my %clientsDesc;

  my $client;
  foreach $client (@clients) {
    my @domains = PleskStructure::getDomainsForClient($client);
    $clientsDesc{$client} = \@domains;
  }

  createClientsDump($root, \%clientsDesc);
}

# $clients - ref to the hash clientNames to ref of array of domainNames
sub createClientsDump {
  my ($root, $clients) = @_;

  my $client;
  my $domainsCount = 0;
  $domainsCount += scalar(@{$clients->{$_}}) foreach keys %{$clients};

  $objDumpStatus->{'COUNTS'}->(scalar(keys %{$clients}), $domainsCount);

  foreach $client (keys %{$clients}) {
    $root->{'ADDCHILD'}->(makeClientNode($client, $clients->{$client}));
  }
}

# $domains - ref to the array of domain names
sub createDomainsDump {
  my ($root, $domains) = @_;

  $objDumpStatus->{'COUNTS'}->(0, scalar(@{$domains}));

  my $domain;
  foreach $domain (@{$domains}) {
    $root->{'ADDCHILD'}->(makeDomainNode($domain));
  }
}

# Condition is not necessary if getting admin parameters
sub getMaxButtonLength {
  my ($parent, $table, $condition) = @_;

  my $sql = "SELECT val FROM $table WHERE param='max_button_length'";
  $sql .= " AND $condition" if $condition;

  if ($wrapDbh->{'EXECUTE'}->($sql) and my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	if ($ptrRow->[0]) {
		$parent->{'ATTRIBUTE'}->('max-button-length', $ptrRow->[0]);
	}
  }
  $wrapDbh->{'FINISH'}->();
}

#Condition is not necessary if getting admin parameters
sub getSkin {
  my ($parent, $table, $condition) = @_;

  return unless PleskVersion::atLeast(5, 0, 0);

  my ($skinId);

  if ($table eq 'dom_param' and !PleskVersion::atLeast(7,5,0)) {
    $skinId = "user_skin_id";
  } else {
    $skinId = "skin_id";
  }

  my $sql = "SELECT s.name FROM Skins as s, $table as t " .
	"WHERE t.param='$skinId' AND t.val = s.id";
  $sql .= " AND $condition" if $condition;

  if ($wrapDbh->{'EXECUTE'}->($sql) and my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	$parent->{'ATTRIBUTE'}->('skin', $ptrRow->[0]);
  }
  $wrapDbh->{'FINISH'}->();
}

sub getLockScreen {
  my ($parent, $table, $condition) = @_;

  return unless PleskVersion::atLeast(7, 5, 2);

  my $sql = "SELECT val FROM $table WHERE param='disable_lock_screen'";
  $sql .= " AND $condition" if $condition;

  my $ptrRow;
  if ($wrapDbh->{'EXECUTE'}->($sql)
	  and $ptrRow = $wrapDbh->{'FETCHROW'}->()
	  and $ptrRow->[0] eq 'true') {
	$parent->{'ATTRIBUTE'}->('lock-screen', 'false');
  }
  $wrapDbh->{'FINISH'}->();
}

#
# Converts the old country codes to the recent ones
#
sub normalizeCountry {
  my ($country) = @_;
  return 'RU' if $country eq 'SU';
  return 'GB' if $country eq 'UK';
  return $country;
}

#
# sub makeClientNode
#
#  arguments:
#         $clientId - ID of client
#         $full - dump all domains
#         $ptrDomainHash
#

sub makeClientNode {
  my($clientName, $domains)=@_;

  my ($item,$sql,$ptrHash,$value,%client,$ptrRow,$id,%clientParams);

  printToLog("Client '$clientName' is started");
  $objDumpStatus->{'ACCOUNT'}->($clientName);
  $objDumpStatus->{'PRINT'}->();

  $sql = "SELECT * FROM clients WHERE login = '$clientName'";

  $wrapDbh->{'EXECUTE'}->($sql);
  %client = %{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();

  my $clientId = $client{'id'};

  $client{'country'} = normalizeCountry($client{'country'});

  my $root = XmlNode('client', 'attributes' => {'id' => $clientId});

  my($attribute,$field,$name);

  if(exists $client{'locale'} && $client{'locale'}) {
    while (($name, $field) = each(%locale_map)) {
      if($client{'locale'} eq $name) {
        $client{'locale'} = $field;
      }
    }
  }

  while (($attribute,$field)=each(%clientsAttribute)){
    if(exists($client{$field})){
	  unless (defined($client{$field})) {
		$client{$field} = '';
	  }
      $root->{'ATTRIBUTE'}->($attribute,$client{$field});
    }
  }

  if (PleskVersion::atLeast(7, 1, 5)) {
	$sql = "SELECT password, type FROM accounts WHERE id = '"
	  . $client{'account_id'} . "'";

	$wrapDbh->{'EXECUTE'}->($sql);
	$ptrHash = $wrapDbh->{'FETCHHASH'}->();
	$item = makePasswordNode($ptrHash->{'password'}, pwdType($ptrHash->{'type'}));
	$wrapDbh->{'FINISH'}->();
  } else {
	$item = makePasswordNode($client{'passwd'},'plain');
  }

  $root->{'ADDCHILD'}->($item);

  my $status = 0;
  if (PleskVersion::atLeast(5, 1, 0)) {
	$status = $client{'status'};
  } else {
	$status |= $Status::ADMIN if $client{'status'} eq 'false';
  }
  $root->{'ADDCHILD'}->(Status::make($status));

  unless (exists $ptrArgs->{'no-content'}) {

	while (($name, $field) = each(%clientsInfo)) {
	  if (exists($client{$field}) && $client{$field}) {
		$item = makeXmlNode('pinfo', $client{$field});
		$item->{'ATTRIBUTE'}->('name', $name);
		$root->{'ADDCHILD'}->($item);
	  }
	}

	# Locale changed format since 7.5.0, and we
	# don't bother converting older formats
	if (PleskVersion::atLeast(7, 5, 0)) {
	  if (exists($client{'locale'}) && $client{'locale'}) {
		my $localeNode = makeXmlNode('pinfo', $client{'locale'});
		$localeNode->{'ATTRIBUTE'}->('name', 'locale');
		$root->{'ADDCHILD'}->($localeNode);
	  }
	}

	unless (PleskVersion::atLeast(6, 0, 0)) {
	  $sql = "SELECT param,val FROM cl_param WHERE cl_id = $clientId";
	  if($wrapDbh->{'EXECUTE'}->($sql)){
		while ($ptrRow=$wrapDbh->{'FETCHROW'}->()){
		  $clientParams{$ptrRow->[0]}=$ptrRow->[1];
		}
	  }
	  $wrapDbh->{'FINISH'}->();
	}

	# button length
	getMaxButtonLength($root, "cl_param", "cl_id=$clientId");
	getSkin($root, "cl_param", "cl_id=$clientId");
	getLockScreen($root, "cl_param", "cl_id=$clientId");

#
# limits
#
	if (PleskVersion::atLeast(6, 0, 0)) {
	  addClientLimits($root, $client{'limits_id'});
	} elsif (PleskVersion::atLeast(5, 0, 0)) {
	  addClientLimits($root, $clientParams{'lim_id'});
	} else {
	  addClientLimits25($root, $clientId);
	}

#
# end limits
#


	if (PleskVersion::atLeast(6, 0, 0)) {
	  addPermissions($root, $client{'perm_id'});
	} elsif (PleskVersion::atLeast(5, 0, 0)) {
	  addPermissions($root, $clientParams{'perm_id'});
	} else {
	  addPermissions25($root, $clientId);
	}

        #fix bug 88139
        if ((PleskVersion::atLeast(7,5,1) and !$client{'perm_id'}) or !PleskVersion::atLeast(7,5,1)) {
          addPermissionNode($root,'cp_access', 'true');
        }

        #fix bug 88544
        if (!PleskVersion::atLeast(7,5,4)) {
          addPermissionNode($root,'manage_not_chroot_shell', 'true');
        }

        if (!PleskVersion::atLeast(8,0,0)) {
          addPermissionNode($root,'stdgui', 'true');
        }

	addClientMulitpleLoginPermission($root, $clientId);
#
# Domain skeleton
#
        if (PleskVersion::atLeast(6, 0, 0)) {
	  my $skeletonPath = AgentConfig::get('HTTPD_VHOSTS_D') . "/.skel/$clientId";

	  my $dumpFile = makeDumpFile("$dumpDir/$client{'login'}.skel", $skeletonPath);
	  if ($dumpFile) {
	    $root->{'ATTRIBUTE'}->('cid-skeleton', $dumpFile);
	  }
        }
  }


  my @ips = PleskStructure::getClientIps($clientName);

  my $ip_pool = makeXmlNode('ip_pool');
  $root->{'ADDCHILD'}->($ip_pool);

  foreach my $ip (@ips) {
    $ip_pool->{'ADDCHILD'}->(PleskStructure::makeIpNode($ip));
  }

  # Site Applications pool
  dumpSiteAppPool($root, $clientId);

  dumpDomainTemplates($root, $clientId);

#  my @domains = PleskStructure::getDomainsForClient($clientName);
#  for my $domainName (@domains) {
#    $root->{'ADDCHILD'}->(makeDomainNode($domainName));
#  }
  for my $domainName (@{$domains}) {
    $root->{'ADDCHILD'}->(makeDomainNode($domainName));
  }

  $root->{'ADDCHILD'}->($_) foreach getCustomButtonsByOwner("client", $clientId);

  return $root;
}

#-----------------------------------------------------------------
#	Dumps information about client's site applications pool
#-----------------------------------------------------------------
sub dumpSiteAppPool( $ $ )
{
	my ($root, $cl_id) = @_;

	# There were no commercial site applications prior to Plesk 7.1.0
	if (!PleskVersion::atLeast(7, 1, 0)) {
		return;
	}

	my $sql = "SELECT sp.name,sp.version,sp.release "
		. "FROM clients c INNER JOIN Repository r ON c.sapp_pool_id=r.rep_id "
		. "LEFT JOIN SiteAppPackages sp ON r.component_id=sp.id "
		. "WHERE c.id='$cl_id' AND (sp.access_level & 1) <> 0";

	if ($wrapDbh->{'EXECUTE'}->($sql)) {

		my $sapp_pool = makeXmlNode('sapp-pool');
		$root->{'ADDCHILD'}->($sapp_pool);

		while (my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {

			my $sapp_package = makeXmlNode('sapp-package');
			$sapp_package->{'ADDCHILD'}->(makeXmlNode('sapp-name', $ptrRow->[0]));

			# version and release may missing
			if ($ptrRow->[1]) {
				$sapp_package->{'ADDCHILD'}->(makeXmlNode('sapp-version', $ptrRow->[1]));
			}
			if ($ptrRow->[2]) {
				$sapp_package->{'ADDCHILD'}->(makeXmlNode('sapp-release', $ptrRow->[2]));
			}

			$sapp_pool->{'ADDCHILD'}->($sapp_package);
		}
	}
	$wrapDbh->{'FINISH'}->();
}


sub addPermissions25 {
  my ($root, $id) = @_;

  # [42098] migration from 7.14 to 7.50 fails
  return unless ($id);

  my ($perm_name, $sql, $ptrRow, $value, $item);
  my @perm_names = ('create_domains', 'manage_dns', 'manage_log');

  foreach $perm_name (@perm_names) {
	$sql = "SELECT DISTINCT val FROM cl_param WHERE cl_id = '$id' AND param = '$perm_name'";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
		my $value = @{$wrapDbh->{'FETCHROW'}->()}[0];
		addPermissionNode($root, $perm_name, $value);
	}
	$wrapDbh->{'FINISH'}->();
  }

  $sql = "SELECT DISTINCT val FROM cl_param WHERE cl_id='$id' AND (param='ip_based_allow' OR param='nb_allow')";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	my $value = @{$wrapDbh->{'FETCHROW'}->()}[0];
	addPermissionNode($root, 'manage_phosting', $value);
  }
  $wrapDbh->{'FINISH'}->();
}

sub addPermissions {
  my ($root, $id) = @_;
  my ($sql, $ptrRow, $item, $name, $value);

  # [42098] migration from 7.14 to 7.50 fails
  return unless ($id);

  $sql = "SELECT permission,value FROM Permissions WHERE id=$id";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  $name = $ptrRow->[0];
	  $value = $ptrRow->[1];
	  next unless exists $xmlPermissions{$name};
	  addPermissionNode($root, $name, $value);
	}
  }
  $wrapDbh->{'FINISH'}->();

  $sql = "SELECT value FROM Permissions WHERE id = $id and ("
	. "permission = 'ipb_allow' or permission = 'nb_allow')";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	my $value = @{$wrapDbh->{'FETCHROW'}->()}[0];
	addPermissionNode($root, 'manage_phosting', $value);
  }
  $wrapDbh->{'FINISH'}->();
}

sub addDomainUserMultipleLoginPermission {
  my ($root, $id) = @_;

  my $sql = "SELECT val FROM dom_param WHERE dom_id=$id AND param='multiply_login' AND val IN ('true','false')";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	addPermissionNode($root, 'multiple-sessions', @{$wrapDbh->{'FETCHROW'}->()}[0]);
  }
  $wrapDbh->{'FINISH'}->();
}

sub addClientMulitpleLoginPermission {
  my ($root, $id) = @_;

  my $sql = "SELECT val FROM cl_param WHERE cl_id =$id AND param='multiply_login'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	addPermissionNode($root, 'multiple-sessions', @{$wrapDbh->{'FETCHROW'}->()}[0]);
  }
  $wrapDbh->{'FINISH'}->();
}

sub addMailUserMultipleLoginPermission {
  my ($root, $id) = @_;

  return unless PleskVersion::atLeast(6, 0, 0);

  my $sql = "SELECT if(val=1,'true',val) FROM mn_param WHERE mn_id=$id AND param='multiply_login'"; # if(val=1.. is a workaround for bug 94747
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	addMailuserPermissionNode($root, 'multiple-sessions', @{$wrapDbh->{'FETCHROW'}->()}[0]);
  }
  $wrapDbh->{'FINISH'}->();
}

sub insertLimitNode {
  my($root,$name,$value) = @_;
  my $item = &makeXmlNode('limit', $value);
  $item->{'ATTRIBUTE'}->('name', $name);
  $root->{'ADDCHILD'}->($item);
}

sub insertLimit {
  my($root,$name,$value) = @_;
  my($item);

  if($value eq '' || $value eq '-1') {
	insertLimitNode($root, $name, '-1');
	return;
  }

  if($name eq 'expiration'){
	my($mday,$mon,$year)=(localtime($value))[3..5];
	$mon++;
	$year+=1900;
	$value = sprintf('%04d-%02d-%02d',$year,$mon,$mday);
  }

  if (!PleskVersion::atLeast(7, 0, 0) && $name eq 'mbox_quota') {
	$value = $value * 1024;
  }

  insertLimitNode($root, $name, $value);
}

sub addClientLimits {
  my ($root, $clientId) = @_;
  return addLimits($root, $clientId, \%clientLimits);
}

sub addClientLimits25 {
  my ($root, $clientId) = @_;
  return addLimits25($root, $clientId, \%clientLimits25, 'cl_id', 'cl_param');
}

sub addDomainLimits {
  my ($root, $clientId) = @_;
  return addLimits($root, $clientId, \%domainLimits);
}

sub addDomainLimits25 {
  my ($root,$domainId) = @_;
  addLimits25($root, $domainId, \%domainLimits25, 'dom_id', 'dom_param');
  #fix bug 50361
  my $ptrRow;
  my $sql ="SELECT traffic FROM hosting WHERE dom_id='$domainId'";
  if ($wrapDbh->{'EXECUTE'}->($sql) && ($ptrRow = $wrapDbh->{'FETCHROW'}->())){
  	insertLimit($root, 'max_traffic', $ptrRow->[0]);
	}
	$wrapDbh->{'FINISH'}->();
      $sql = "SELECT size FROM domains WHERE id='$domainId'";
  if ($wrapDbh->{'EXECUTE'}->($sql) && ($ptrRow = $wrapDbh->{'FETCHROW'}->())){

  	insertLimit($root, 'disk_space', $ptrRow->[0]);

	}
	$wrapDbh->{'FINISH'}->();
}

sub addLimits25 {
  my($root, $limitId, $limitNames, $idName, $paramName)=@_;

  if(!$limitId) {
	return;
  }

  my($value, $name, $count, $sql, $ptrRow);

  $sql = "SELECT val, param FROM $paramName WHERE $idName='$limitId'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  ($name, $value) = @{$ptrRow};
	}
  }

  foreach $name (%{$limitNames}) {
	$sql = "SELECT val FROM $paramName WHERE $idName = '$limitId' AND param = '$name'";

	if(($wrapDbh->{'EXECUTE'}->($sql)) && ($ptrRow = $wrapDbh->{'FETCHROW'}->())) {
	  ($value) = @{$ptrRow};
	  if($name eq 'max_pbox') {
		$name = 'max_box';
	  }
	  insertLimit($root, $name, $value);
	}
	$wrapDbh->{'FINISH'}->();
  }
}

#
# sub addLimits
#
# arguments:
#           $root    - XML node to add 'limit' node
#           $limitID - limit id
#
# return:
#           $count - number of the added nodes
#
sub addLimits {
  my ($root, $limitId, $limitNames) = @_;

  my $count;

  if (!$limitId) {
    foreach my $key (keys %{$limitNames}) {
      next if $key eq 'max_pbox';
      insertLimit($root, $key, "-1");
      $count++;
    }
    return $count;
  }

  my ($sql, $ptrRow, $value, $name);
  $sql = "SELECT limit_name,value FROM Limits WHERE id=$limitId";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
    $count=0;
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
      ($name, $value) = @{$ptrRow};
      next unless exists $limitNames->{$name};
	  insertLimit($root, $name, $value);
      $count++;
    }
  }
  $wrapDbh->{'FINISH'}->();
  return $count;
}

#
# Gather domain templates information
#
sub dumpDomainTemplates {
  my ($root, $clientId) = @_;

  return unless PleskVersion::atLeast(5, 0, 0);

  my $sql;
  if (PleskVersion::atLeast(6, 0 ,0)) {
	$sql = "SELECT tmpl_id FROM clients WHERE id=$clientId";
  } else {
	$sql = "SELECT val FROM cl_param WHERE cl_id=$clientId AND ".
	  "param = 'dom_tmpl_list_id'";
  }

  my $tmpl_id;
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	$tmpl_id = @{$wrapDbh->{'FETCHROW'}->()}[0];
  }
  $wrapDbh->{'FINISH'}->();
  return unless $tmpl_id;

  my %templateNames;
  my %templateNotes;
  $sql = "SELECT t.id, t.name, t.note_id FROM Templates t, Repository r "
	. "WHERE r.rep_id=$tmpl_id AND r.component_id = t.id";
  $wrapDbh->{'EXECUTE'}->($sql);
  while (my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	$templateNames{$ptrRow->[0]} = $ptrRow->[1];
	$templateNotes{$ptrRow->[0]} = $ptrRow->[2];
  }
  $wrapDbh->{'FINISH'}->();

  while (my ($tmpl_id, $tmpl_name) = each %templateNames) {
	my $node = makeXmlNode('domain-template');
	$node->{'ATTRIBUTE'}->('name', $tmpl_name);
	if ($templateNotes{$tmpl_id}) {
	  $node->{'ATTRIBUTE'}->('note-id', $templateNotes{$tmpl_id});
	}

	my $logrotation_id;
	my $sql = "SELECT element, value FROM TmplData WHERE tmpl_id=$tmpl_id";
	$wrapDbh->{'EXECUTE'}->($sql);
	while (my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  my ($element, $value) = @{$ptrRow};

	  if(!defined($value)) { $value = ''; }

	  $value = -1 if ($element eq 'stat_ttl' and $value == 0
					  and !PleskVersion::atLeast(6, 0, 0));

	  $value = -1 if ($element eq 'disk_space'
					  or $element eq 'max_traffic'
					  or $element eq 'mbox_quota') and $value == 0;

	  $value = 'physical' if ($element eq 'vh_type' and $value ne 'none' and
							  !PleskVersion::atLeast(6, 0, 0));

	  if ($element eq 'logrotation_id') {
		$logrotation_id = $value;
		next;
	  }

	  next if not exists $template_items{$element};

	  $value*= 1024 if ($element eq 'mbox_quota' and $value != -1 and !PleskVersion::atLeast(7, 0, 0));
	  $value*= 1024*1024 if ($element eq 'quota' and $value != -1 and !PleskVersion::atLeast(7, 0, 0));

	  if ($element eq 'quota' and $value == '0') {
		$value = -1;
	  }

	  if ($value eq '' and exists $emptyableBoolElementInTemplate{$element}) {
		$value = 'false';
	  }

	  addTemplateItem($node, $element, $value);
	}
	$wrapDbh->{'FINISH'}->();

	if ($logrotation_id) {
	  makeLogrotationNode($node, $logrotation_id);
	}

	unless (PleskVersion::atLeast(6, 0, 0)) {
	  addTemplateItem($node, 'dns_type', 'master');
	}

	$root->{'ADDCHILD'}->($node);
  }
}

sub addTemplateItem {
  my ($node, $name, $value) = @_;
  my $child = makeXmlNode('template-item', $value);
  $child->{'ATTRIBUTE'}->('name', $name);
  $node->{'ADDCHILD'}->($child);
}

#
# Returns the XML node if domain service is present and undef
# otherwise
#

sub getDomainServiceStatus {
  my ($domId, $service) = @_;

  my $status = 0;
  my $sql = "SELECT status FROM DomainServices "
	. " WHERE dom_id=$domId AND type='$service'";

  if ($wrapDbh->{'EXECUTE'}->($sql)	and my $ptrRow = $wrapDbh->{'FETCHROW'}->()){
	$status = $ptrRow->[0];
  } else {
    $status = undef;
  }
  $wrapDbh->{'FINISH'}->();

  if (defined($status)) {
    return Status::make($status);
  }
}

#
# makeDomainNode
#
# arguments:
#           $domainId - domain id
#
# return:
#           $root - pointer to XML node
#
sub makeDomainNode {
  my($domainName)=@_;

  printToLog("Domain '$domainName' is started");

  $objDumpStatus->{'DOMAIN'}->($domainName);
  $objDumpStatus->{'PRINT'}->();

  my($sql,%domain,%domParams,$ptrRow,$ptrHash,$id);
  my $only_mail = exists $ptrArgs->{'only-mail'};

#
# get domain's info
#
  my $domainNameColumn = 'name';
  if (PleskVersion::atLeast(7, 1, 0)) {
    $domainNameColumn = 'displayName';
  }

  printToLog("Getting domain info");
  $sql = "SELECT * FROM domains WHERE $domainNameColumn = '$domainName'";

  unless($wrapDbh->{'EXECUTE'}->($sql)){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }
  unless($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }
  %domain=%{$ptrHash};
  $wrapDbh->{'FINISH'}->();

  my $domainId = $domain{'id'};

  $sql = "SELECT param,val FROM dom_param WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
      $domParams{$ptrRow->[0]}=$ptrRow->[1];
    }
  }
  $wrapDbh->{'FINISH'}->();

#
# end get domain's info
#

  if (PleskVersion::atLeast(7, 1, 0)) {
	$domainName = $domain{'displayName'};
  } else {
	$domainName = $domain{'name'};
  }

  my $root = XmlNode('domain',
                     'attributes' => {'name' => $domainName,
                                      'id' => $domainId});

  my $domainAsciiName = $domain{'name'};
  my $domainType = $domain{'htype'};

  addWwwStatus($root, $domainId, $domainAsciiName);

  # Domain's IP address
  printToLog("Getting domain IP");
  my $ip = PleskStructure::getDomainIp($domainName);
  if ($ip) {
    $root->{'ADDCHILD'}->(PleskStructure::makeIpNode($ip));
  }

  #
  # Status
  #
  printToLog("Dupming domain status");

  my $status = 0;
  if (PleskVersion::atLeast(5, 1, 0)) {
	$status = $domain{'status'};
  } else {
	$status |= $Status::ADMIN if $domain{'admin_status'} eq 'false';
	$status |= $Status::CLIENT if $domain{'status'} eq 'false';
  }
  $root->{'ADDCHILD'}->(Status::make($status));

  #
  # No further info required if 'no-content' specified
  #
  if (exists $ptrArgs->{'no-content'} && !$only_mail) {

	# Need to dump information about all databases for DbServers mapping

		printToLog("Dumping domain databases");

		$sql = "SELECT id FROM data_bases WHERE dom_id=$domainId";
		if ($wrapDbh->{'EXECUTE'}->($sql)) {

			my (@databases);
			while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
				push @databases,$ptrRow->[0];
			}
			$wrapDbh->{'FINISH'}->();

			foreach my $dbid (@databases) {
				my ($item) = makeDatabaseNode($dbid);
				$root->{'ADDCHILD'}->($item) if ref($item) =~ /HASH/;
			}
		}
		$wrapDbh->{'FINISH'}->();
		return $root;
  }

  # Check whether this domain is default on IP
  if (PleskVersion::atLeast(6, 0, 0)) {
	$sql = "SELECT COUNT(*) FROM IP_Addresses WHERE default_domain_id=$domainId";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
	  if (@{$wrapDbh->{'FETCHROW'}->()}[0] ne "0") {
		$root->{'ADDCHILD'}->(makeXmlNode('default-domain'));
	  }
	}
	$wrapDbh->{'FINISH'}->();
  } else {
    if (PleskStructure::isExclusiveIp($ip)) {
	  $root->{'ADDCHILD'}->(makeXmlNode('default-domain'));
	}
  }

  printToLog("Getting domain limits");
#
# limits
#
  if (!$only_mail) {
    if (PleskVersion::atLeast(6, 0, 0)) {
      addDomainLimits($root,$domain{'limits_id'});
    } elsif (PleskVersion::atLeast(5, 0, 0)) {
      addDomainLimits($root,$domParams{'lim_id'});
    } else {
      addDomainLimits25($root,$domainId);
    }
  }

#
# end limits
#

#
# permissions
#

# bug 79536 
#  if ($id = $domParams{'perm_id'}) {
#    addPermissions($root,$id);
#  }

#
# end permissions
#

  printToLog("Dumping domain mailsystem");

  my $mailServiceStatus;
  if (PleskVersion::atLeast(6, 0, 0)) {
    $mailServiceStatus = getDomainServiceStatus($domainId, 'mail');
  } else {
    $mailServiceStatus = Status::make($Status::ENABLED);
  }

  if ($mailServiceStatus) {
    my $rootMail = makeXmlNode('mailsystem');
    $rootMail->{'ADDCHILD'}->($mailServiceStatus);

    if (PleskVersion::atLeast(6, 0, 0)) {
      $sql = "SELECT m.id, a.password, a.type FROM mail m".
        " LEFT JOIN accounts a ON m.account_id=a.id ".
        " WHERE m.dom_id=$domainId ORDER BY m.mail_name";
    } else {
      $sql = "SELECT id, password, 'plain' FROM mail ".
        " WHERE dom_id=$domainId ORDER BY mail_name";
    }

    if ($wrapDbh->{'EXECUTE'}->($sql)){
      my(@mails);
      while($ptrRow = $wrapDbh->{'FETCHROW'}->()){
        push @mails,[@{$ptrRow}];
      }
      $wrapDbh->{'FINISH'}->();

      foreach $ptrRow (@mails){
        $ptrRow->[2] = pwdType($ptrRow->[2]);
        my $item = makeMailUserNode(@{$ptrRow},$domainName, $domainAsciiName);
        $rootMail->{'ADDCHILD'}->($item) if ref($item) =~ /HASH/;
      }
    } else {
      $wrapDbh->{'FINISH'}->();
    }

    getCatchAllAddress($rootMail, $domainId);

    $root->{'ADDCHILD'}->($rootMail);
  }

	my @SiteApplications;

	if (!$only_mail) {
		# Domain's DNS settings
		$root->{'ADDCHILD'}->(makeDomainDnsZone(\%domain));

		# Domain's aliases
		dumpDomainAliases($domainId, $root);

		#-----------------------------------------------------------------
		# Information about domain's site applications
		# should be extracted prior to database/hosting/custom-buttons
		# dump, as site applications' linked resources should be dumped
		# separately.
		#-----------------------------------------------------------------
		@SiteApplications = getDomainSiteapps($domainId);

		#-----------------------------------------------------------------
		# Domain's databases
		#-----------------------------------------------------------------
		printToLog("Dumping domain databases");

		# Site applications' databases should not be dumped here
		my $exclude_list = '';
		{
			my @excluded_dbs;
			foreach my $sapp (@SiteApplications) {
				push @excluded_dbs, $sapp->getDatabases();
			}
			if (@excluded_dbs) {
				$exclude_list = " AND id NOT IN(" . getSqlList(@excluded_dbs) . ")";
			}
		}

		# Databases without those used in site apps
		$sql = "SELECT id FROM data_bases WHERE dom_id=$domainId" . $exclude_list;
		if ($wrapDbh->{'EXECUTE'}->($sql)) {

			my (@databases);
			while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
				push @databases,$ptrRow->[0];
			}
			$wrapDbh->{'FINISH'}->();

			foreach my $dbid (@databases) {
				my ($item) = makeDatabaseNode($dbid);
				$root->{'ADDCHILD'}->($item) if ref($item) =~ /HASH/;
			}
		}
		$wrapDbh->{'FINISH'}->();
	}

#
# maillist
#
  printToLog("Dumping domain maillists");
  if (PleskVersion::atLeast(6, 0, 0)) {

    my $maillistsStatus = getDomainServiceStatus($domainId, 'maillists');
    if ($maillistsStatus) {
      my $mlNode = makeXmlNode('maillists');
      $mlNode->{'ADDCHILD'}->($maillistsStatus);

      $sql="SELECT id,name,status,dom_id FROM MailLists WHERE dom_id=$domainId";
      if($wrapDbh->{'EXECUTE'}->($sql)){
        my (@mlists);
        while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
          push @mlists,[@{$ptrRow}];
        }
        $wrapDbh->{'FINISH'}->();
        foreach $ptrRow (@mlists){
          my $item = makeMailListNode(@{$ptrRow});
          if(ref($item)=~/HASH/){
            $mlNode->{'ADDCHILD'}->($item);
          }
        }
      }
      $wrapDbh->{'FINISH'}->();

      $root->{'ADDCHILD'}->($mlNode);
    }
  }
#
# end maillist
#

	if ($only_mail) {
		# end dump
		return $root;
	}

  printToLog("Dumping domain statistics");
#
# traffic
#
  &addCurrentTraffic($root,$domainId);
#
# end traffic
#


  printToLog("Dumping domain certificates");
#
# certificates
#
  if (PleskVersion::atLeast(6, 0, 0)) {
	  my $cert_id = $domain{'cert_rep_id'};
	  $cert_id = '' unless (defined($cert_id));
    $sql = "SELECT c.id FROM certificates c, Repository r ".
      "WHERE c.id=r.component_id AND r.rep_id='$cert_id' ORDER BY c.id";
  } else {
    $sql = "SELECT id FROM certificates WHERE dom_id=$domainId";
  }
  my @ids;
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
      push @ids, $ptrRow->[0];
    }
  }
  $wrapDbh->{'FINISH'}->();

  foreach $id (@ids) {
	my $item = makeCertificateNode($id);
	if (ref($item)=~/HASH/) {
	  $root->{'ADDCHILD'}->($item);
	}
  }
#
# end certificates
#

#
# tomcat service
#
  printToLog("Dumping Tomcat service");

  my $tomcatServiceStatus;
  if (PleskVersion::atLeast(6, 0, 0)) {
    $tomcatServiceStatus = getDomainServiceStatus($domainId, 'tomcat');
  }

  if ($tomcatServiceStatus) {
	my $tomcatNode = makeXmlNode('tomcat');
	$tomcatNode->{'ADDCHILD'}->($tomcatServiceStatus);

	$sql = "SELECT wa.name,wa.status FROM WebApps wa LEFT JOIN DomainServices ds ON wa.domain_service_id=ds.id WHERE ds.dom_id = $domainId";

	my (%webapps,$webapp);

	if ($wrapDbh->{'EXECUTE'}->($sql)) {

	  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		$webapps{$ptrRow->[0]}	= $ptrRow->[1];
	  }
	}
	$wrapDbh->{'FINISH'}->();

	foreach $webapp (keys %webapps) {
	  makeWebAppNode($tomcatNode,$webapp,$webapps{$webapp},$domainAsciiName);
	}

	$root->{'ADDCHILD'}->($tomcatNode);
  }
#
# end tomcat service
#

  printToLog("Dumping domain user");
#
# domainuser
#
  my $item = makeDomainUserNode($domainId);
  if(ref($item)=~/HASH/){
		printToLog("Getting domain permissions"); # bug 79536 

		if ($id = $domParams{'perm_id'}) {
			addPermissions($item,$id);
		}

		$root->{'ADDCHILD'}->($item);
  }

#
# end domainuser
#

  printToLog("Dumping domain hosting");
#
# hosting
#
  my $hosting;
  if($domainType=~/vrt_hst/){
    $hosting = makePhostingNode(\%domain, \@SiteApplications);

    if(defined $domParams{'wu_script'} && $domParams{'wu_script'} eq 'true') {
    	$hosting->{'ATTRIBUTE'}->('wu_script', 'true');
	}
  }elsif($domainType=~/std_fwd/){
    $hosting = makeShostingNode(\%domain);
  }elsif($domainType=~/frm_fwd/){
    $hosting = makeFhostingNode(\%domain);
  }
  if(ref($hosting)=~/HASH/){
    $root->{'ADDCHILD'}->($hosting);
  }

#
# end hosting
#

  $root->{'ADDCHILD'}->($_)
	foreach getCustomButtonsByOwner('domain-admin', $domainId);

  return $root;
}

#-----------------------------------------------------------------
#	Returns <dns> or <dns-zone> xml elements
#-----------------------------------------------------------------
sub makeDomainDnsZone( $ )
{
	my ($domainHashPtr) = @_;

	printToLog("Dumping domain DNS");

	if (PleskVersion::atLeast(8, 0, 0)) {
		return makeDnsZone($domainHashPtr->{'dns_zone_id'});
	} else {
		return makeDnsZoneOld($domainHashPtr->{'id'});
	}
}

#-----------------------------------------------------------------
#	Returns <dns-zone-param> xml node
#-----------------------------------------------------------------
sub makeDnsZoneParam( $ $ $ )
{
	my ($name, $db_unit, $db_value) = @_;

	my $dnsZoneParam = makeXmlNode('dns-zone-param');
	$dnsZoneParam->{'ATTRIBUTE'}->('name',  $name);
	$dnsZoneParam->{'ATTRIBUTE'}->('value', $db_value);

	my %units = (
		'second'  => 1,
		'minutes' => 60,
		'hours'   => 3600,
		'days'    => 86400,
		'weeks'   => 604800
	);
	my $string_unit;

	foreach my $quant (keys %units) {
		if ($units{$quant} == $db_unit) {
			$string_unit = $quant;
			last;
		}
	}
	$dnsZoneParam->{'ATTRIBUTE'}->('unit', $string_unit);

	return $dnsZoneParam;
}

#-----------------------------------------------------------------
#	Returns <dnsrec> xml node or false if given record should not be dumped
#-----------------------------------------------------------------
sub makeDnsRecord( $ )
{
	my ($ptrHash) = @_;

	return if ($ptrHash->{'type'} =~ /none/);

	if ($ptrHash->{'type'} eq 'TXT') {
		$ptrHash->{'val'} =~ s/"(.*)"/$1/;
	}

	my ($item) = makeXmlNode('dnsrec');
	$item->{'ATTRIBUTE'}->('type', $ptrHash->{'type'});
	$item->{'ATTRIBUTE'}->('src', $ptrHash->{'host'});
	$item->{'ATTRIBUTE'}->('dst', $ptrHash->{'val'});
	$item->{'ATTRIBUTE'}->('opt', $ptrHash->{'opt'}) if ($ptrHash->{'opt'});

	return $item;
}

sub makeDnsZone( $ )
{
	my ($dnsZoneId) = @_;

	my $sql = "SELECT * FROM dns_zone WHERE id=$dnsZoneId";

    if (!$wrapDbh->{'EXECUTE'}->($sql)) {
      $wrapDbh->{'FINISH'}->();
      Logging::error("Broken referencial integrity: DNS zone id $dnsZoneId is not found in dns_zone");
      return;
    }

    my $ptrHash = $wrapDbh->{'FETCHHASH'}->();

    my $email;
    if (length($ptrHash->{'email'}) != 0) {
      $email = $ptrHash->{'email'};
    } else {
      # should assume some default email or something...
      $email = 'root@localhost.localdomain';
    }

    my $dnsZone = XmlNode('dns-zone',
                          'attributes' => {'email' => $email, 'type' => $ptrHash->{'type'}},
                          'children' => [Status::make($ptrHash->{'status'})]);

    foreach my $zoneParam ('ttl', 'refresh', 'retry', 'expire', 'minimum') {
      $dnsZone->{'ADDCHILD'}->(
                               makeDnsZoneParam($zoneParam,
                                                $ptrHash->{$zoneParam . '_unit'},
                                                $ptrHash->{$zoneParam}));
    }

	$wrapDbh->{'FINISH'}->();

	# dns records
	$sql = "SELECT * FROM dns_recs WHERE dns_zone_id=$dnsZoneId";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {

		while (my $ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
			my $dnsrec = makeDnsRecord($ptrHash);
			if ($dnsrec) {
				$dnsZone->{'ADDCHILD'}->($dnsrec);
			}
		}
	}
	$wrapDbh->{'FINISH'}->();

	return $dnsZone;
}

sub makeDnsZoneOld( $ )
{
	my ($domainId) = @_;

	my $dnsNode = makeXmlNode('dns');

	my $sql = "SELECT dns_zone";
	if (PleskVersion::atLeast(5, 0, 0)) {
		$sql .= ",dns_type"
	}
	$sql .= " FROM domains WHERE id='$domainId'";

	if ($wrapDbh->{'EXECUTE'}->($sql) and my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		$dnsNode->{'ATTRIBUTE'}->('enabled', $ptrRow->[0] eq 'false' ? 'false' : 'true');
		$dnsNode->{'ATTRIBUTE'}->('type', (!$ptrRow->[1] or ($ptrRow->[1] eq 'master')) ? 'master' : 'slave');
	}
	$wrapDbh->{'FINISH'}->();

	$sql = "SELECT * FROM dns_recs WHERE dom_id=$domainId";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
      my $ptrHash;
		while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
			my $dnsrec = makeDnsRecord($ptrHash);
			if ($dnsrec) {
				$dnsNode->{'ADDCHILD'}->($dnsrec);
			}
          $ptrHash = undef;
		}
	}
	$wrapDbh->{'FINISH'}->();

        # Add old master record

        $sql = "SELECT * FROM dns_masters WHERE dom_id = '$domainId'";

        if ($wrapDbh->{'EXECUTE'}->($sql)) {
           while (my $ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
             my $oldMasterRec = makeOldMasterRec($ptrHash);
             if ($oldMasterRec){
               $dnsNode->{'ADDCHILD'}->($oldMasterRec);
             }
             $ptrHash = undef;
           }
        }
        $wrapDbh->{'FINISH'}->();
	return $dnsNode;
}

sub makeOldMasterRec
{
        my ($ptrHash) = @_;
        my $item = XmlNode('dnsrec');
        $item->{'ATTRIBUTE'}->('src',$ptrHash->{'ip_address'});
        $item->{'ATTRIBUTE'}->('type', 'master');
        return $item;
}
#-----------------------------------------------------------------
#	Dumps domain aliases for given domain ID
#-----------------------------------------------------------------
sub dumpDomainAliases( $ $ )
{
	my ($domainId, $root) = @_;

	# Domain aliases are introduced since Plesk 7.5.5
	if (!PleskVersion::atLeast(8, 0, 0)) {
		return;
	}

	printToLog('Dumping domain aliases... ', 1);

	my @aliases;

	my $sql = "SELECT * FROM domainaliases WHERE dom_id='$domainId'";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {

		while (my $ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
			push @aliases, $ptrHash;
		}
	}
	$wrapDbh->{'FINISH'}->();

	foreach my $alias (@aliases) {

		my $aliasNode = XmlNode('domain-alias',
                                'attributes' => {'name' => $alias->{'displayName'},
                                                 'mail' => $alias->{'mail'},
                                                 'web' => $alias->{'web'}},
                                'children' => [Status::make($alias->{'status'})]);

		# DNS zone
		if ($alias->{'dns_zone_id'} != 0) {
			$aliasNode->{'ADDCHILD'}->(makeDnsZone($alias->{'dns_zone_id'}));
		}

		$root->{'ADDCHILD'}->($aliasNode);
	}

	printToLog('OK');
}

#-----------------------------------------------------------------
#	Returns array of SiteApp objects, representing all site applications
#	installed on domain and its subdomains.
#-----------------------------------------------------------------
sub getDomainSiteapps( $ )
{
	my ($dom_id) = @_;

	# Site applications migration is supported since Plesk 7.0.0
	if (!PleskVersion::atLeast(7, 0, 0)) {
		return;
	}

	my @SiteApplications;
	my @sapp_ids;

	# ids of site apps installed on domain
	my $sql = "SELECT id FROM SiteApps WHERE dom_type='domain' AND dom_id='$dom_id'";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
		while (my $idRow = $wrapDbh->{'FETCHROW'}->()) {
			push @sapp_ids, $idRow->[0];
		}
	}
	$wrapDbh->{'FINISH'}->();

	# subdomains
	$sql = "SELECT id FROM subdomains WHERE dom_id=$dom_id";
	my (@subdoms);
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
		while (my $subdomRow = $wrapDbh->{'FETCHROW'}->()) {
			push @subdoms, $subdomRow->[0];
		}
	}
	$wrapDbh->{'FINISH'}->();

	# ids of site apps installed on subdomains
	foreach my $subdom_id (@subdoms) {
		$sql = "SELECT id FROM SiteApps WHERE dom_type='subdomain' AND dom_id='$subdom_id'";
		if ($wrapDbh->{'EXECUTE'}->($sql)) {
			while (my $idRow = $wrapDbh->{'FETCHROW'}->()) {
				push @sapp_ids, $idRow->[0];
			}
		}
        $wrapDbh->{'FINISH'}->();
	}

	# create SiteApp objects
	foreach my $sapp_id (@sapp_ids) {
		my $sapp = new SiteApp($sapp_id, $wrapDbh);
		if ($sapp->error()) {
			printToError($sapp->error());
			next;
		}
		push @SiteApplications, $sapp;
	}

	return @SiteApplications;
}


sub addWwwStatus {
  my ($parent, $domainId, $domainName) = @_;

  my $sql;

  if (PleskVersion::atLeast(8, 0, 0)) {
	$sql = "SELECT r.* FROM dns_recs r, domains d WHERE d.id=$domainId "
		 . "AND d.dns_zone_id=r.dns_zone_id AND r.type = 'CNAME' AND r.host = 'www.$domainName.'";
  } else {
  	$sql = "SELECT * FROM dns_recs WHERE dom_id = $domainId AND type = 'CNAME' " .
	"AND host = 'www.$domainName.'";
  }

  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	$parent->{'ATTRIBUTE'}->('www', 'true');
  } else {
	$parent->{'ATTRIBUTE'}->('www', 'false');
  }
  $wrapDbh->{'FINISH'}->();
}

sub makeDomainUserNode {
  my($domainId)=@_;

  my($sql,$root,$item,$ptrHash,$xmlName,$fieldName);

  $sql="SELECT * FROM dom_level_usrs WHERE dom_id=$domainId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }
  my %domainUser=%{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();
  $root = &makeXmlNode('domainuser');

  if ($domainUser{'state'}) {
	$root->{'ATTRIBUTE'}->('status', $domainUser{'state'});
  }

  if (PleskVersion::atLeast(7, 1, 5)) {
    if ($domainUser{'account_id'} != 0) {
      $root->{'ADDCHILD'}->(makeAccountPasswordNode($domainUser{'account_id'}));
    }
  } else {
	$root->{'ADDCHILD'}->(makePasswordNode($domainUser{'passwd'}, 'plain'));
  }

  if (PleskVersion::atLeast(5, 0, 0) && defined($domainUser{'card_id'})) {
	$sql = "SELECT * from Cards WHERE id=$domainUser{'card_id'}";
	if($wrapDbh->{'EXECUTE'}->($sql)){
	  $ptrHash = $wrapDbh->{'FETCHHASH'}->();
	  while (($xmlName,$fieldName)=each %domainUserAttribute){
		if($ptrHash->{$fieldName}){
		  $root->{'ATTRIBUTE'}->($xmlName,$ptrHash->{$fieldName});
		}
	  }

	  while (($xmlName, $fieldName) = each %domainUserInfo) {
		my $value = $ptrHash->{$fieldName};
		if ($value) {
		  $value = normalizeCountry($value) if $fieldName eq 'country';
		  $item = makeXmlNode('pinfo', $value);
		  $item->{'ATTRIBUTE'}->('name', $xmlName);
		  $root->{'ADDCHILD'}->($item);
		}
	  }
	}
	$wrapDbh->{'FINISH'}->();
  }

  addDomainUserLocale($root,$domainId);

  if (PleskVersion::atLeast(7, 1, 0)) {
	addPermissions($root, $domainUser{'perm_id'});
  }
  addDomainUserMultipleLoginPermission($root, $domainId);

  if (!PleskVersion::atLeast(8,0,0)) {
    addPermissionNode($root,'stdgui', 'true');
  }

  # max button length
  getMaxButtonLength($root, "dom_param", "dom_id=$domainId");
  getSkin($root, "dom_param", "dom_id=$domainId");
  getLockScreen($root, "dom_param", "dom_id=$domainId");

  return $root;
}

sub addDomainUserLocale {
  my ($root,$domainId)=@_;
  my ($locale,$ptrRow,$sql,$item,$key,$value);

  if (PleskVersion::atLeast(7,1,0)){
    $locale = "locale";
  } else {
    $locale = "user_locale";
  }

  $sql = "SELECT val FROM dom_param WHERE dom_id = $domainId AND param = '$locale'";
  if ($wrapDbh->{'EXECUTE'}->($sql) && ($ptrRow = $wrapDbh->{'FETCHROW'}->())) {
    if($ptrRow->[0]){
      while (($key,$value) = each(%locale_map)){
        if ($ptrRow->[0] eq $key) {
          $ptrRow->[0] = $value;
        }
      }
      $item = makeXmlNode('pinfo', $ptrRow->[0]);
      $item->{'ATTRIBUTE'}->('name', 'locale');
      $root->{'ADDCHILD'}->($item);
    }
  }
  $wrapDbh->{'FINISH'}->();
}
#
# addCurrentTraffic - add current traffic
#
#     arguments:
#                 $root - XML node to add traffic's nodes
#                 $domainId - ID of domain
#
sub addCurrentTraffic {
  my ($root, $domainId) = @_;
  my ($sql, $ptrRow, $item);

  my $trafficValue = '';

  if (PleskVersion::atLeast(6, 0, 0)) {
	$sql = "SELECT http_in, http_out, ftp_in, ftp_out, smtp_in, smtp_out, ".
	  "pop3_imap_in, pop3_imap_out, date FROM DomainsTraffic WHERE dom_id=$domainId";

    if ($wrapDbh->{'EXECUTE'}->($sql)) {
      my @keys = ('http', 'ftp', 'smtp', 'pop3-imap');
      my ($key,$i);
      while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		for($i = 0; $i < @keys*2; ++$i) {
		  if($ptrRow->[$i]) {
			$trafficValue .= $ptrRow->[8];
			$trafficValue .= ' ';
			$trafficValue .= $keys[$i/2];
			$trafficValue .= ' ';
			$trafficValue .= ($i%2 ? 'out' : 'in');
			$trafficValue .= ' ';
			$trafficValue .= $ptrRow->[$i];
			$trafficValue .= "\n";
		  }
		}
      }
    }
	$wrapDbh->{'FINISH'}->();
  } else {
	$sql = "SELECT transfer, date FROM stat WHERE dom_id=$domainId";
    if ($wrapDbh->{'EXECUTE'}->($sql)) {
      while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		if ($ptrRow->[0]) {
		  $trafficValue .= $ptrRow->[1];
		  $trafficValue .= ' http in ';
		  $trafficValue .= $ptrRow->[0];
		  $trafficValue .= "\n";
		}
	  }
    }
	$wrapDbh->{'FINISH'}->();
  }

  $root->{'ADDCHILD'}->(makeXmlNode('traffic', $trafficValue));
}


# There are 3 types of nonexistent user mail handling:
# Bounce with message (bounce:message text)
# Catch to address (email@address)
# SMTP reject (reject)
# Returns bounce|catch|reject or empty string
sub getNonexistentMode
{
	my ($domainId) = @_;

	my $sql = "SELECT p.value FROM Parameters p, DomainServices ds " .
	"WHERE ds.parameters_id = p.id AND ds.dom_id = $domainId AND p.parameter = 'nonexist_mail'";
	my $mode;
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
      $mode = @{$wrapDbh->{'FETCHROW'}->()}[0];
	} else {
      $mode = "bounce:This address no longer accepts mail.";
    }
	$wrapDbh->{'FINISH'}->();

	return $mode;
}

sub getCatchAllAddress
{
	my ($parent, $domainId) = @_;

	my $sql;
	if (PleskVersion::atLeast(6, 0, 0)) {
		my $mode = getNonexistentMode($domainId);

		if ($mode =~ /^catch$/) {
			$sql = "SELECT p.value FROM Parameters p, DomainServices ds " .
			"WHERE ds.parameters_id = p.id AND ds.dom_id = $domainId AND p.parameter = 'catch_addr'";

		} elsif ($mode =~ /^bounce$/) {
			$sql = "SELECT CONCAT('bounce:', p.value) FROM Parameters p, DomainServices ds " .
			"WHERE ds.parameters_id = p.id AND ds.dom_id = $domainId AND p.parameter = 'bounce_mess'";

		} elsif ($mode =~ /^reject$/) {
			$sql = "SELECT 'reject'";
		}

	} elsif (PleskVersion::atLeast(5, 0, 0)) {
		$sql = "SELECT p.value FROM Parameters p, dom_param dp WHERE " .
		"dp.dom_id = $domainId AND dp.param='param_id' AND dp.val = p.id AND p.parameter = 'catch_addr'";

	} else {
		$sql = "SELECT dp.val FROM dom_param dp WHERE dp.dom_id = $domainId AND dp.param = 'catch_addr'";
	}

	# some not supported mode or default parameter [bug 43901]
	return unless ($sql);

	if ($wrapDbh->{'EXECUTE'}->($sql)) {
		my $catchAllAddr = @{$wrapDbh->{'FETCHROW'}->()}[0];
		if ($catchAllAddr) {
			$parent->{'ADDCHILD'}->(makeXmlNode('catch-all', $catchAllAddr));
		}
	}
	$wrapDbh->{'FINISH'}->();
}

#
# makeMailListNode
#
#   arguments:
#              $mlistId - ID of the mail list
#              $mlistName - name
#              $mllistState - state of maillist
#              $domId - domain Id. Due to bug 96558
#
#   return:
#              $root - XML node
#
sub makeMailListNode {
  my($mlistId,$mlistName,$mlistState, $domId)=@_;
  my($cmd,$item);

  unless (defined Mailman::version()) {
    printToLog("Unable to found Mailman installation");
    return;
  }
  my $sql = "SELECT displayName FROM domains WHERE id = $domId";
  $wrapDbh->{'EXECUTE'}->($sql);
  my $ptrRow = $wrapDbh->{'FETCHROW'}->();
  my $domName = $ptrRow->[0];
  $wrapDbh->{'FINISH'}->();

  my $root = XmlNode('maillist',
                     'attributes'=>{'name', $mlistName},
                     'children' => [Status::make($mlistState)]);

  for my $listOwner (Mailman::getListOwners($mlistName)) {
    $item = XmlNode('owner', 'content' => $listOwner);
    $root->{'ADDCHILD'}->($item);
  }

  unless(ref($item)=~/HASH/){
    $item = XmlNode('owner', 'content' => $mlistName."@".$domName);
    $root->{'ADDCHILD'}->($item);
  }

  $root->{'ADDCHILD'}->(makePasswordNode(Mailman::getListPassword($mlistName)), 'encrypted');

  my %listMembers = Mailman::getListMembers($mlistName);
  for my $memberEmail (keys %listMembers) {
    my $member = XmlNode('recipient', 'content' => $memberEmail);
    if ($listMembers{$memberEmail} ne "") {
      $member->{'ATTRIBUTE'}->('fullname', $listMembers{$memberEmail});
    }
    $root->{'ADDCHILD'}->($member);
  }

  my $archiveDir = AgentConfig::get("MAILMAN_VAR_D") . "/archives/private/$mlistName.mbox";
  my $archive = "$mlistName.mbox";

  if (-d $archiveDir and -f "$archiveDir/$archive") {
	my $cid = makeDumpFile("$dumpDir/$mlistName.mlist",
						   $archiveDir,
						   $archive, undef,
						   "-h");
    $root->{'ATTRIBUTE'}->('cid', $cid) if ($cid);
  }

  return $root;
}


sub addUrlDecodedTextNode {
  my ($parent, $name, $value) = @_;

  $parent->{'ADDCHILD'}->(makeXmlNode($name, urlDecode($value))) if ($value);
}

sub makeCertificateNode {
  my($certId)=@_;

  my($sql,$root,%cert,$item);

  $sql = "SELECT * FROM certificates WHERE id=$certId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    &printToError("Error: makeCertificateNode: certificate ID '$certId' is not found");
    return undef;
  }

  %cert= %{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();

  $root = makeXmlNode('certificate');

  if (PleskVersion::atLeast(6, 0, 0)) {
	addUrlDecodedTextNode($root, 'certificate-data', $cert{'cert'});
	addUrlDecodedTextNode($root, 'signing-request', $cert{'csr'});
	addUrlDecodedTextNode($root, 'ca-certificate', $cert{'ca_cert'});
  } elsif (PleskVersion::atLeast(5, 0, 0)) {
	addUrlDecodedTextNode($root, 'certificate-data', $cert{'pub_key'});
  }
  addUrlDecodedTextNode($root, 'private-key', $cert{'pvt_key'});
  if (PleskVersion::atLeast(6, 0, 0)) {
	  $root->{'ATTRIBUTE'}->('name', $cert{'name'});
  } else {
	  # certificate names were not used in Plesk prior to 6.0
	  $root->{'ATTRIBUTE'}->('name', $certId);
  }

  return $root;
}

sub makeDatabaseNode {
  my ($dbId) = @_;
  my ($ptrHash, $ptrRow, $item, $charset, $dbServerHost);

  my $sql = "SELECT name, type FROM data_bases WHERE id=$dbId";
  unless ($wrapDbh->{'EXECUTE'}->($sql)) {
    $wrapDbh->{'FINISH'}->();
    return;
  }
  my ($dbName, $dbTypePlesk) = @{$wrapDbh->{'FETCHROW'}->()};
  $wrapDbh->{'FINISH'}->();

  if ($dbTypePlesk =~ /postgres/) {
	$dbTypePlesk = "postgresql";
  } elsif ($dbTypePlesk ne "mysql") {
    printToError("Unknown database type: $dbTypePlesk");
    return;
  }

  my $root = XmlNode('database', 'attributes' => {'name' => $dbName, 'type' => $dbTypePlesk});

  # Remote databases hosting was introduced in 8.0.0
  if (PleskVersion::atLeast(8, 0, 0)) {
    my $sql = "SELECT host,port,ds.type FROM DatabaseServers as ds, data_bases as db WHERE "
      . "ds.id = db.db_server_id AND db.id = $dbId";
    $wrapDbh->{'EXECUTE'}->($sql);

	$ptrRow = $wrapDbh->{'FETCHROW'}->();

	$dbServerHost = ($ptrRow->[0] ne 'localhost') ? $ptrRow->[0] : 'localhost';
	my $dbServerNode = XmlNode('db-server');
	$dbServerNode->{'ATTRIBUTE'}->('type', $ptrRow->[2]);
	$dbServerNode->{'ADDCHILD'}->(XmlNode('host', 'content' => "$ptrRow->[0]"));
	$dbServerNode->{'ADDCHILD'}->(XmlNode('port', 'content' => "$ptrRow->[1]"));

    $wrapDbh->{'FINISH'}->();

	$root->{'ADDCHILD'}->($dbServerNode);

  }

  if (exists $ptrArgs->{'no-content'}){
    return $root;
  }

  printToLog("Database $dbName");

  if (PleskVersion::atLeast(7, 5, 0)) {
    $sql = "SELECT login, password, type FROM db_users, accounts WHERE "
      . "db_users.db_id = $dbId AND db_users.account_id = accounts.id";

    $wrapDbh->{'EXECUTE'}->($sql);
    while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
      my $dbUserRoot = makeXmlNode('dbuser');
      $dbUserRoot->{'ATTRIBUTE'}->('name',$ptrHash->{'login'});

      $item = makePasswordNode($ptrHash->{'password'}, pwdType($ptrHash->{'type'}));

      $dbUserRoot->{'ADDCHILD'}->($item);
      $root->{'ADDCHILD'}->($dbUserRoot);
    }
  } else {
    $wrapDbh->{'EXECUTE'}->("SELECT login, passwd FROM db_users WHERE db_id = $dbId");

    while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
      my $dbPasswdType;

      # In Plesk 7.1.5 DB password type wasn't described in the DB. We
      # can distinguish the type only using length of password.
      if (length($ptrHash->{'passwd'}) > 14) {
        $dbPasswdType = 'encrypted';
      } else {
        $dbPasswdType = 'plain';
      }

      my $dbUserRoot = makeXmlNode('dbuser');
      $dbUserRoot->{'ATTRIBUTE'}->('name', $ptrHash->{'login'});

      $item = makePasswordNode($ptrHash->{'passwd'}, $dbPasswdType);

      $dbUserRoot->{'ADDCHILD'}->($item);
      $root->{'ADDCHILD'}->($dbUserRoot);
    }
  }
  $wrapDbh->{'FINISH'}->();

  my ($dbType, $dbUser, $dbPasswd, $pg_manage);

  if ($dbTypePlesk eq "mysql") {
    $dbType = "mysql";
  } elsif ($dbTypePlesk eq "postgresql") {
    $dbType = "Pg";
  }

  # remote database hosting regularizes the database connectivity even for local hosts.
  # thanks goes to xman.
  if (PleskVersion::atLeast(8, 0, 0)) {
    my $sql = "SELECT admin_login, admin_password FROM DatabaseServers as ds, data_bases as db "
      . "WHERE ds.id = db.db_server_id AND db.id = $dbId";
    $wrapDbh->{'EXECUTE'}->($sql);
    ($dbUser, $dbPasswd) = @{$wrapDbh->{'FETCHROW'}->()};
    $wrapDbh->{'FINISH'}->();

    if ($dbTypePlesk =~ /mysql/ and $dbServerHost eq 'localhost') {
      $dbPasswd = AgentConfig::get('password');
    }

    if ($dbTypePlesk=~/postgresql/) {
      $root->{'ATTRIBUTE'}->('version',getPostgresVersion());
    }
  } else {
    if ($dbTypePlesk=~/mysql/) {
      $dbUser = $wrapDbh->{'USER'}->();
      $dbPasswd = $wrapDbh->{'PASSWORD'}->();
    } elsif ($dbTypePlesk =~ /postgresql/) {
      $sql = "SELECT param, val FROM misc WHERE param like 'postgresql_admin_\%'";
      if ($wrapDbh->{'EXECUTE'}->($sql)) {
        while ($ptrRow=$wrapDbh->{'FETCHROW'}->()) {
          my ($name, $value) = @{$ptrRow};
          $dbUser = $value if $name =~ /login$/;
          $dbPasswd = $value if $name =~ /passwd$/;
        }
      }
      $wrapDbh->{'FINISH'}->();

      $pg_manage = AgentConfig::get("PRODUCT_ROOT_D") . "/admin/sbin/pg_manage";
    }
  }

  if (my $outFile = makeDumpDb($dbName, $dbType, $dbUser, $dbPasswd, $dbServerHost, undef, undef, undef, $pg_manage)) {
    $root->{'ATTRIBUTE'}->('cid',$outFile);
  }

  if ($dbTypePlesk=~/postgresql/) {
     my $psql = $bin{'psql'};

     $charset = `export PGUSER=$dbUser PGPASSWORD='$dbPasswd'; $psql -l | grep '^[ \t]*$dbName ' | awk '{print \$5}'`;
     chomp($charset);
     if($charset ne '') { $root->{'ATTRIBUTE'}->('charset', $charset); }
  }

  if ($dbTypePlesk=~/mysql/) {
    $root->{'ATTRIBUTE'}->('version', getMysqlVersion(\%bin));
  }

  return $root;
}

#
# makeMailUserNode
#
# argumets:
#           $mailId - mail ID
#           $passwd - password
#           $typePasswd - type of password
#           $domainName - name of domain
#
# return:
#           $root = XML node 'mailuser'
#
sub makeMailUserNode {
  my ($mailId, $passwd, $typePasswd, $domainName, $domainAsciiName) = @_;

  my($sql,%mail,$item,$ptrRow,$ptrHash,$dir,$id, $mbox_quota);

  my $dumpFile;

  $sql = "SELECT * FROM mail WHERE id = $mailId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    printToError("Error: makeMailUserNode: mail ID '$mailId' is not found");
    return undef;
  }
  %mail = %{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();
  my $mailName = $mail{'mail_name'};

  my $root = XmlNode('mailuser', 'attributes' => {'id' => $mailId,
                                                 'name' => $mailName});

  if (defined($passwd) && $passwd) {
	if ($passwd =~ /NULL/) {
	  $item = makePasswordNode('', 'plain');
	} else {
	  $item = makePasswordNode($passwd,$typePasswd);
	}

    $root->{'ADDCHILD'}->($item);
  }
	# bug 78844
	if ($mail{'perm_id'}) {
		$sql = "SELECT value FROM Permissions WHERE id = $mail{'perm_id'} AND permission = 'cp_access'";
		unless($wrapDbh->{'EXECUTE'}->($sql)){
			printToError("Error: makeMailUserNode: unable to get cp-access permission for mail ID '$mailId'");
	 	}
 		$ptrRow = $wrapDbh->{'FETCHROW'}->();
	}
	else {
		$ptrRow->[0] = $mail{'cp_access'};
	}

        if ($ptrRow->[0]){
	  addMailuserPermissionNode($root, 'cp-access', $ptrRow->[0]);
        }
	$wrapDbh->{'FINISH'}->();

  # multiple login
  addMailUserMultipleLoginPermission($root, $mailId);


  if($mail{'mbox_quota'}){
	$mbox_quota = $mail{'mbox_quota'};

	if (!PleskVersion::atLeast(6, 0, 0) and $mbox_quota ne "-1") {
	  $mbox_quota = $mbox_quota * 1024;
	}

	$root->{'ATTRIBUTE'}->('mailbox-quota',$mbox_quota);
  }



  my $enable_mailbox = $mail{'postbox'} =~ /true/;

  # Check whether there autoresponder with attach
  # On 'Olde Pleskes' there was bug allowing attaches
  # when mailbox in turned off, so we have to explicitly
  # turn mailbox on if there is attach.
  $sql = "SELECT COUNT(ra.filename) FROM mail_resp AS mr, resp_attach as ra
            WHERE ra.rn_id = mr.id AND mr.mn_id = $mailId";
  if ($wrapDbh->{'EXECUTE'}->($sql)
	  and ($ptrRow = $wrapDbh->{'FETCHROW'}->())
	  and $ptrRow->[0] != 0) {
	$enable_mailbox = 1;
  }
  $wrapDbh->{'FINISH'}->();

  if ($enable_mailbox) {
    $item = makeXmlNode('mailbox');
    $item->{'ATTRIBUTE'}->('type','mdir');
    $root->{'ADDCHILD'}->($item);

	if ($mail{'postbox'} =~ /true/) {
	  $item->{'ATTRIBUTE'}->('enabled', 'true');
	} else {
	  $item->{'ATTRIBUTE'}->('enabled', 'false');
	}

    $dir=AgentConfig::get('QMAIL_MAILNAMES_D')."/$domainName/$mailName/Maildir";
    $dumpFile = &makeDumpFile("$dumpDir/$mailName\@$domainName.mdir",
			     $dir);
    if($dumpFile){
      $item->{'ATTRIBUTE'}->('cid',$dumpFile);
    }
  }

#
# aliases
#
  if (PleskVersion::atLeast(5, 0, 0)) {
	$sql="SELECT alias FROM mail_aliases WHERE mn_id=$mailId";
	if($wrapDbh->{'EXECUTE'}->($sql)){
	  while ($ptrRow=  $wrapDbh->{'FETCHROW'}->()){
		$item=&makeXmlNode('alias',$ptrRow->[0]);
		$root->{'ADDCHILD'}->($item);
	  }
	}
	$wrapDbh->{'FINISH'}->();
  }
#
# end aliases
#

#
# redirect
#
  if ($mail{'redir_addr'}) {
    $item = makeXmlNode('redirect',$mail{'redir_addr'});
	$item->{'ATTRIBUTE'}->('enabled', $mail{'redirect'});
    $root->{'ADDCHILD'}->($item);
  }
#
# end redirect
#

#
# mail group
#
  $root->{'ATTRIBUTE'}->('mailgroup-enabled', $mail{'mail_group'});

  $sql="SELECT address FROM mail_redir WHERE mn_id=$mailId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
	while ($ptrRow=  $wrapDbh->{'FETCHROW'}->()){
      if($ptrRow->[0] ne $mailName . "@" . $domainName) {
	    $item=&makeXmlNode('mailgroup-member', $ptrRow->[0]);
	    $root->{'ADDCHILD'}->($item);
      }
	}
  }
  $wrapDbh->{'FINISH'}->();

#
# end mail group
#

#
# autoresponders
#
  my $autorespondersNode = makeXmlNode('autoresponders');
  $autorespondersNode->{'ATTRIBUTE'}->('enabled', $mail{'autoresponder'});

  $dir=AgentConfig::get('QMAIL_MAILNAMES_D')."/$domainName/$mailName/\@attachments";
  $dumpFile = &makeDumpFile("$dumpDir/$mailName\@$domainName.attach",
							$dir);
  if($dumpFile){
	$autorespondersNode->{'ATTRIBUTE'}->('cid_attaches',$dumpFile);
  }

  my(@autos);
  $sql = "SELECT id FROM mail_resp WHERE mn_id=$mailId ORDER BY id";
  if($wrapDbh->{'EXECUTE'}->($sql)){
	while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
	  push @autos,$ptrRow->[0];
	}
  }
  $wrapDbh->{'FINISH'}->();

  my %attaches;
  my @autoresponders;

  foreach $id (@autos) {
	my $item = &makeAutoresponderNode($id, $mailName . "@" . $domainName);

	push @autoresponders, $item;
  }

  my $util = AgentConfig::get("PRODUCT_ROOT_D") . "/admin/sbin/";
  if (PleskVersion::atLeast(6, 5, 0)) {
 	$util .= "mailmng --list-attachments --domain-name=$domainName --mailname=$mailName";
  } else {
 	$util .= "list_files -1A '" . AgentConfig::get("QMAIL_MAILNAMES_D") . "/$domainName/$mailName/\@attachments/'";
  }

  my @filelist;
  open(LIST, "$util |");
  while (<LIST>) {
 	chomp;
 	my ($file) = split /\t/;
 	push @filelist, $file;
  }
  #clear list if utility failed
  unless (close(LIST)) {
	@filelist = ();
  }

  # Old Plesk utilities printed return code on stdout (#66905)
  unless (PleskVersion::atLeast(7, 1, 0)) {
 	if ($#filelist != -1 && $filelist[$#filelist] =~ /^\d$/) {
 		$#filelist -= 1;
 	}
  }
  unless(exists $ptrArgs->{'configuration-only'}) {
    foreach my $file (@filelist) {
  	  my $attachNode = makeXmlNode('attach');
	  $attachNode->{'ATTRIBUTE'}->('file', $file);
	  $autorespondersNode->{'ADDCHILD'}->($attachNode);
    }
  }

  foreach my $autoresponder (@autoresponders) {
	$autorespondersNode->{'ADDCHILD'}->($autoresponder);
  }
  $root->{'ADDCHILD'}->($autorespondersNode);
#
# end autoresponders
#

#
# Addressbook
#
  if (PleskVersion::atLeast(6, 0, 0)) {
	makeAddressbookNode($root, $mailName . '@' . $domainName);
  }

#
# Spamassassin
#
  if (PleskVersion::atLeast(7, 0, 0)) {
  	makeSpamassassinNode($root,$mail{'mail_name'}.'@'.$domainAsciiName);
  }
#
# End Spamassassin
#

#
# Drweb
#
if (-e "/etc/drweb/users.conf"){
  	$item=&makeDrwebNode($root,$mailName.'@'.$domainName);
  }
#
# End Drweb
#

  if (PleskVersion::atLeast(6, 0, 0)) {
	getMaxButtonLength($root, "mn_param", "mn_id=$mailId");
	getSkin($root, "mn_param", "mn_id=$mailId");
	getLockScreen($root, "mn_param", "mn_id=$mailId");
  }

  $root->{'ADDCHILD'}->($_) foreach getCustomButtonsByOwner('mailuser', $mailId);

  return $root;
}

#
# makeAutoresponderNode
#
#  arguments:
#             $id - ID of autoreponder
#
#  return:
#             $root - XML node
#
sub makeAutoresponderNode {
  my($autoId, $mailName)=@_;

  my($name,$value,$sql,%auto,$ptrRow,$item);

  $sql = "SELECT * FROM mail_resp WHERE id=$autoId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    &printToError("Error: makeAutoresponderNode: autoresponder # $autoId is not found");
    return undef;
  }

  %auto=%{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();

  my $root = &makeXmlNode('autoresponder');

  $value=$wrapBase64->{'ENCODE'}->($auto{'text'});
  $item=&makeXmlNode('text',$value);
  if (exists $auto{'charset'} && $auto{'charset'} ne '') {
	$item->{'ATTRIBUTE'}->('charset', $auto{'charset'});
  }
  $root->{'ADDCHILD'}->($item);

  if (exists $auto{'content_type'} && $auto{'content_type'}) {
	$root->{'ATTRIBUTE'}->('content-type', $auto{'content_type'});
  }

  $name = $auto{'resp_name'};
  $root->{'ATTRIBUTE'}->('name',$name);

  if ($auto{'resp_on'}=~/false/){
    $root->{'ATTRIBUTE'}->('status','off');
  }

  unless ($auto{'key_where'}=~/no/){
    $value=$auto{'key_where'}.':'.$auto{'keystr'};
    $root->{'ATTRIBUTE'}->('require',$value);
  }

  if($auto{'subject'}){
    $value = $wrapBase64->{'ENCODE'}->($auto{'subject'});
    chomp $value;
    $root->{'ATTRIBUTE'}->('subject',$value);
  }

  if ($auto{'reply_to'} and $auto{'reply_to'} ne $mailName) {
	$root->{'ATTRIBUTE'}->('replyto',$auto{'reply_to'});
  }

#
# forward
#
  $sql="SELECT address FROM resp_forward WHERE rn_id=$autoId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    my (@list);
    while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
	  # skip empty entries - found somewhere @ the wild Net
      push @list, $ptrRow->[0] if $ptrRow->[0];
    }
	if (@list) {
	  $root->{'ATTRIBUTE'}->('redirect',join(',',@list));
	}
  }
  $wrapDbh->{'FINISH'}->();
#
# end forward
#
  $item = &makeXmlNode('autoresponder-limit',$auto{'ans_freq'});
  $item->{'ATTRIBUTE'}->('name','ans-freq');
  $root->{'ADDCHILD'}->($item);

  $item = &makeXmlNode('autoresponder-limit',$auto{'mem_limit'});
  $item->{'ATTRIBUTE'}->('name','ans-count');
  $root->{'ADDCHILD'}->($item);


  my @attaches;
#
# attachment
#
  $sql = "SELECT filename FROM resp_attach WHERE rn_id=$autoId";
  if($wrapDbh->{'EXECUTE'}->($sql) and !exists $ptrArgs->{'configuration-only'}){

    while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
      $item=&makeXmlNode('attach');
      $item->{'ATTRIBUTE'}->('file',$ptrRow->[0]);
      $root->{'ADDCHILD'}->($item);
    }
  }
  $wrapDbh->{'FINISH'}->();
#
# end attachment
#

  return $root;
}


sub makeAddressbookNode {
  my ($parent, $mailName) = @_;

  my $sql = "SELECT Db FROM mysql.db WHERE db='horde'";
  return unless $wrapDbh->{'EXECUTE'}->($sql);
  $wrapDbh->{'FINISH'}->();

  $sql = "SHOW TABLES FROM horde LIKE 'turba_objects'";
  return unless $wrapDbh->{'EXECUTE'}->($sql);
  $wrapDbh->{'FINISH'}->();

  my $node = makeXmlNode('addressbook');

  my %objectFields = ('id'=>'id',
					  'name'=>'name',
					  'alias'=>'alias',
					  'email'=>'email',
					  'title'=>'title',
					  'company'=>'company',
					  'home-phone'=>'homephone',
					  'work-phone'=>'workphone',
					  'mobile-phone'=>'cellphone',
					  'fax'=>'fax',
					  'home-address'=>'homeaddress',
					  'work-address'=>'workaddress',
					  'notes'=>'notes');

  my $ptrHash;
  $sql = "SELECT * FROM horde.turba_objects WHERE owner_id = '$mailName'";
  $wrapDbh->{'EXECUTE'}->($sql);

  while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
	if ($ptrHash->{'object_type'} eq 'Group') {
	  my $group = makeXmlNode('addressbook-group');
	  $group->{'ATTRIBUTE'}->('id', $ptrHash->{'object_id'});

	  my $members = $ptrHash->{'object_members'};
	  while ($members =~ /"(.*?)"/g) {
		my $member = $1;
		chomp $member;

		my $memberNode = makeXmlNode('addressbook-member');
		$memberNode->{'ATTRIBUTE'}->('id', $member);
		$group->{'ADDCHILD'}->($memberNode);
	  }

	  $node->{'ADDCHILD'}->($group);
	} else {
	  my $user = makeXmlNode('addressbook-contact');
	  my ($field, $attr);
	  while (($field, $attr) = each %objectFields) {
		if (defined $ptrHash->{'object_' . $attr}) {
          # Here 'encoded' attribute means there is no meaning to
          # convert data to UTF as it was already converted manually
		  my $value = $ptrHash->{'object_' . $attr};
		  $value = $wrapBase64->{'ENCODE'}->($value) if ($attr ne 'id');
		  $user->{'ATTRIBUTE'}->($field, $value);
		}
	  }
	  $node->{'ADDCHILD'}->($user);
	}
  }
  $wrapDbh->{'FINISH'}->();

  $parent->{'ADDCHILD'}->($node);
}

sub makePhostingNode {
  my ($ptrDomain, $ptrSiteApplications) = @_;

  my @SiteApplications = @{$ptrSiteApplications};

  unless(ref($ptrDomain)=~/HASH/){
    &printToError("Error: makePhostNode: bag arguments");
    return undef;
  }
  my $root=&makeXmlNode('phosting');

  my($domainName,$domainRoot,$path,$fileName,$sql,%hosting,$domainId,
    $xmlName,$fieldName,$id,$item,$ptrRow,$ptrHash);
  $domainName = $ptrDomain->{'name'};
  $domainId = $ptrDomain->{'id'};

	# shared http/https content
	if (PleskVersion::atLeast(7, 5, 3)) {
		my $sql = "SELECT same_ssl FROM hosting WHERE dom_id=$domainId AND same_ssl='true'";
		if ($wrapDbh->{'EXECUTE'}->($sql)) {
			$root->{'ATTRIBUTE'}->('shared-content', 'true');
		}
		$wrapDbh->{'FINISH'}->();
	}

	#-----------------------------------------------------------------
	# Site applications content should not be included into dump
	# together with other website content.
	#-----------------------------------------------------------------
	my @exclude_htdocs_files;
	my @exclude_htdocs_ssl_files;
	my @exclude_cgibin_files;
	foreach my $sapp (@SiteApplications) {
		if (!$sapp->isDomainSiteapp()) {
			next;
		}
		if ($sapp->isSsl()) {
			push @exclude_htdocs_ssl_files, $sapp->getHtdocsFiles();
		} else {
			push @exclude_htdocs_files, $sapp->getHtdocsFiles();
		}
		push @exclude_cgibin_files, $sapp->getCgibinFiles();
	}

    # Hosting content dump
	$domainRoot = AgentConfig::get("HTTPD_VHOSTS_D") . "/$domainName";

	if (-d $domainRoot) {

		# /httpdocs
		if ($fileName = makeDumpFile("$dumpDir/$domainName.htdocs", "$domainRoot/httpdocs", '.', \@exclude_htdocs_files)) {
			$root->{'ATTRIBUTE'}->('cid_docroot', $fileName);
		}

		# /httpsdocs
		if ($fileName = makeDumpFile("$dumpDir/$domainName.shtdocs", "$domainRoot/httpsdocs", '.', \@exclude_htdocs_ssl_files)) {
			$root->{'ATTRIBUTE'}->('cid_docroot_ssl', $fileName);
		}

		# /cgi-bin
		if ($fileName = makeDumpFile("$dumpDir/$domainName.cgi", "$domainRoot/cgi-bin", '.', \@exclude_cgibin_files)) {
			$root->{'ATTRIBUTE'}->('cid_cgi', $fileName);
		}

		# since Plesk 5.0.0 /statistics/webstat. Before /httpdocs/webstat
               my $webstatRoot = (PleskVersion::atLeast(5,0,0)) ? $domainRoot."/statistics/webstat" : $domainRoot."/httpdocs/webstat";
		if (-d $webstatRoot) {
			if ($fileName = makeDumpFile("$dumpDir/$domainName.webstat", "$webstatRoot")) {
				$root->{'ATTRIBUTE'}->('cid_webstat', $fileName);
			}
		}
		if (-d $domainRoot . "/statistics/webstat-ssl") {
			if ($fileName = makeDumpFile("$dumpDir/$domainName.webstat-ssl", "$domainRoot/statistics/webstat-ssl")) {
				$root->{'ATTRIBUTE'}->('cid_webstat_ssl', $fileName);
			}
		}

		# /statistics/ftpstat
		if (-d $domainRoot . "/statistics/ftpstat") {
			if ($fileName = makeDumpFile("$dumpDir/$domainName.ftpstat", "$domainRoot/statistics/ftpstat")) {
				$root->{'ATTRIBUTE'}->('cid_ftpstat', $fileName);
			}
		}

		# /private
		if (-d $domainRoot . "/private") {
			if ($fileName = makeDumpFile("$dumpDir/$domainName.private", "$domainRoot/private")) {
				$root->{'ATTRIBUTE'}->('cid_private', $fileName);
			}
		}
	}

  $sql = "SELECT * FROM hosting WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)) {
    %hosting = %{$wrapDbh->{'FETCHHASH'}->()};
  }
  $wrapDbh->{'FINISH'}->();

  while (($xmlName,$fieldName) = each(%hostingAttribute)) {
    if (defined($hosting{$fieldName}) and $hosting{$fieldName} eq "true") {
      $root->{'ATTRIBUTE'}->($xmlName, 'true');
    }
  }

  $sql = "SELECT val FROM dom_param WHERE dom_id='$domainId' "
	. "AND param='apacheErrorDocs'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	if ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  my $errdocs = (@{$ptrRow})[0];
	  if (defined($errdocs) and $errdocs eq 'true') {
		$root->{'ATTRIBUTE'}->('errdocs', 'true');
	  }
	}
  }
  $wrapDbh->{'FINISH'}->();

#
# sysuser
#
  Logging::info("Dumping system user");

  if (PleskVersion::atLeast(5, 0, 0)) {
	if($id = $hosting{'sys_user_id'}){
	  $item= &makePleskSysUserNode($id);
	  if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
	  }
	}
  } else {
	$item = &makeSyntheticSysUserNode($hosting{'login'}, $hosting{'passwd'}, 'plain');
	if(ref($item) =~ /HASH/) {
	  $root->{'ADDCHILD'}->($item);
	}
  }
#
# end sysuser
#

#
# scripting
#
  $item = &makeXmlNode('scripting');
  while(($xmlName,$fieldName)=each(%hostingScripting)){
    next if $fieldName=~/asp_dot_net/;
    if(defined($hosting{$fieldName}) and $hosting{$fieldName}=~/true/){
      $item->{'ATTRIBUTE'}->($xmlName,'true');
    }
  }
  $root->{'ADDCHILD'}->($item);
#
# end scripting
#

	#-----------------------------------------------------------------
	# Dump of installed site applications
	#-----------------------------------------------------------------
	foreach my $sapp (@SiteApplications) {
		if (!$sapp->isDomainSiteapp()) {
			next;
		}
		$root->{'ADDCHILD'}->(dumpSiteApplication($sapp, AgentConfig::get("HTTPD_VHOSTS_D") . "/$domainName"));
	}

#
# frontpage user
#
  if (exists $hosting{'fp_adm'} and $hosting{'fp_adm'}) {
    $item = &makeXmlNode('fpuser');
    $item->{'ATTRIBUTE'}->('name',$hosting{'fp_adm'});
    if($hosting{'fp_pass'}){
      my $passNode = &makePasswordNode($hosting{'fp_pass'},'plain');
      $item->{'ADDCHILD'}->($passNode);
    }
    $root->{'ADDCHILD'}->($item);
  } elsif (exists $hosting{'fp'} and $hosting{'fp'} eq 'true' and
  		  !$hosting{'fp_adm'} and !$hosting{'fp_pass'}) {

	# Workaround for older Plesks: during upgrade 1.3 -> 2.0 (or something)
	# frontpage administrator records were lost
	$item = makeXmlNode('fpuser');
	if (PleskVersion::atLeast(5, 0, 0)) {
		my %sysuser = getSysUserInfo($hosting{'sys_user_id'});
		$item->{'ATTRIBUTE'}->('name', lc($sysuser{'login'}));
		$item->{'ADDCHILD'}->(makeAccountPasswordNode($sysuser{'account_id'}));
	} else {
		$item->{'ATTRIBUTE'}->('name', $hosting{'login'});
		$item->{'ADDCHILD'}->(makePasswordNode($hosting{'passwd'}, 'plain'));
	}
	$root->{'ADDCHILD'}->($item);
  }
#
# end frontpage user
#

#
# logrotation
#
  &makeDomainLogrotationNode($root, $domainId);
#
# end logrotation
#

#
# anonftp
#
  &addAnonFtp($root,$domainId,$domainName);
#
# end anonftp
#

#
# protected dirs
#
  if (PleskVersion::atLeast(8, 0, 0)) {
	$sql = "SELECT id,path,realm,non_ssl,`ssl`,cgi_bin FROM protected_dirs WHERE dom_id=$domainId ORDER BY id";
  } elsif (PleskVersion::atLeast(6, 0, 0)) {
	$sql = "SELECT id,path,realm,non_ssl,`ssl` FROM protected_dirs WHERE dom_id=$domainId ORDER BY id";
  } else {
	$sql = "SELECT id,path,realm FROM protected_dirs WHERE dom_id=$domainId ORDER BY id";
  }

  if($wrapDbh->{'EXECUTE'}->($sql)){
    my (@dirs);
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
      push @dirs,[@{$ptrRow}];
    }
    $wrapDbh->{'FINISH'}->();
    foreach $ptrRow (@dirs){
      unless (defined($hosting{'ssl'}) and $hosting{'ssl'}=~/true/){
        $ptrRow->[4] = 'false'; # bug 97159
      }
      $item = &makeProtDirNode(@{$ptrRow});
      if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
      }
    }
  }
  $wrapDbh->{'FINISH'}->();
#
# end protected dirs
#

#
# web users
#
  $sql = "SELECT * FROM web_users WHERE dom_id=$domainId ORDER BY id";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    my (@webs);
    while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
      push @webs,{%{$ptrHash}};
    }
    $wrapDbh->{'FINISH'}->();
    foreach $ptrHash (@webs){
      $item = &makeWebUserNode($ptrHash,$domainName);
      if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
      }
    }
  } else {
    $wrapDbh->{'FINISH'}->();
  }
#
# end web users
#

#
# subdomains
#
  if (PleskVersion::atLeast(7, 0, 0)) {
    $sql = "SELECT * FROM subdomains WHERE dom_id=$domainId ORDER BY id";

	my (@subdoms);
    if($wrapDbh->{'EXECUTE'}->($sql)){
      while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
		push @subdoms,{%{$ptrHash}};
      }
	}
	$wrapDbh->{'FINISH'}->();
	foreach $ptrHash (@subdoms){
	  $item = dumpSubDomain($ptrHash, $domainName, \@SiteApplications);
	  if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
	  }
	}
  }
#
# end subdomains
#

  #
  # configuration files (conf/vhost[_ssl].conf)
  #
  $domainRoot = AgentConfig::get("HTTPD_VHOSTS_D").'/'.$domainName;
  my $vhostConf = $domainRoot .'/conf/vhost.conf';
  my $vhostSslConf = $domainRoot .'/conf/vhost_ssl.conf';
  if (-f $vhostConf || -f $vhostSslConf) {
	if($fileName = makeDumpFile("$dumpDir/$domainName.conf", "$domainRoot/conf", ["vhost.conf","vhost_ssl.conf"])) {
	  $root->{'ATTRIBUTE'}->('cid_conf', $fileName);
	}
  }
  #
  # end configuration files (conf/vhost[_ssl].conf)
  #

  #
  # Webalizer configuration
  #
  if (PleskVersion::atLeast(7, 5, 3)) {
	my $node = makeXmlNode("webalizer");

	my $sql = "SELECT referrer FROM webalizer_hidden_referrer "
	  . " WHERE dom_id=$domainId";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
	  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		if ($ptrRow->[0] eq "Direct Request") {
		  $node->{'ATTRIBUTE'}->('hide-direct-referrer', 'true');
		} else {
		  $node->{'ADDCHILD'}->(makeXmlNode("webalizer-hidden-referrer", $ptrRow->[0]));
		}
	  }
	}
	$wrapDbh->{'FINISH'}->();

	$sql = "SELECT referrer, group_name FROM webalizer_group_referrer "
	  . " WHERE dom_id=$domainId";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
	  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		my $ref = makeXmlNode("webalizer-group-referrer", $ptrRow->[0]);
		$ref->{'ATTRIBUTE'}->('group-name', $ptrRow->[1]);
		$node->{'ADDCHILD'}->($ref);
	  }
	}
	$wrapDbh->{'FINISH'}->();

	$root->{'ADDCHILD'}->($node);
  }

  # Cron


  return $root;
}

#-----------------------------------------------------------------
#	Returns XmlNode of sapp-installed type or false
#-----------------------------------------------------------------
sub dumpSiteApplication( $ $ )
{
	my ($sapp, $vhost_path) = @_;

	my $sapp_installed = makeXmlNode('sapp-installed');

	#-----------------------------------------------------------------
	# sapp-spec
	#-----------------------------------------------------------------
	my $sapp_spec = makeXmlNode('sapp-spec');
	$sapp_spec->{'ADDCHILD'}->(makeXmlNode('sapp-name', $sapp->getName()));
	if ($sapp->getVersion()) {
		$sapp_spec->{'ADDCHILD'}->(makeXmlNode('sapp-version', $sapp->getVersion()));
	}
	if ($sapp->getRelease()) {
		$sapp_spec->{'ADDCHILD'}->(makeXmlNode('sapp-release', $sapp->getRelease()));
	}
	if ($sapp->getDescription()) {
		my $desc = $wrapBase64->{'ENCODE'}->($sapp->getDescription());
		$sapp_spec->{'ADDCHILD'}->(XmlNode('sapp-description', 'attributes' => {'encoding' => 'base64'}, 'content' => $desc));
	}
	if ($sapp->isCommercial()) {
		$sapp_spec->{'ADDCHILD'}->(makeXmlNode('sapp-commercial'));
	}
	if ($sapp->isIntegrated()) {
		$sapp_spec->{'ADDCHILD'}->(makeXmlNode('sapp-integrated'));
	}

	$sapp_installed->{'ADDCHILD'}->($sapp_spec);

	#-----------------------------------------------------------------
	#sapp-param*
	#-----------------------------------------------------------------
	while (my ($param_name, $param_value) = each %{$sapp->getParams()}) {
		my $name  = $wrapBase64->{'ENCODE'}->($param_name);
		my $value = $wrapBase64->{'ENCODE'}->($param_value);

		$sapp_installed->{'ADDCHILD'}->(XmlNode('sapp-param',
				'children' => [XmlNode('sapp-param-name',  'attributes' => {'encoding' => 'base64'}, 'content' => $name),
							   XmlNode('sapp-param-value', 'attributes' => {'encoding' => 'base64'}, 'content' => $value)]));
	}

	#-----------------------------------------------------------------
	# Linked resources
	#-----------------------------------------------------------------

	# Databases
	foreach my $db_id ($sapp->getDatabases()) {
      my $database = makeDatabaseNode($db_id);
      if (ref($database) =~ /HASH/) {
        $sapp_installed->{'ADDCHILD'}->($database);
      }
    }

	# Custom buttons
	foreach my $cb_id ($sapp->getCustomButtons()) {
		my $cb = getCustomButtonById71($cb_id);
		if (ref($cb) =~ /HASH/) {
			$sapp_installed->{'ADDCHILD'}->($cb);
			push @skip_custom_buttons, $cb_id;
		}
	}

	#-----------------------------------------------------------------
	# Site application content
	#-----------------------------------------------------------------
	my $prefix = $sapp->getInstallPrefix();

	# sapp-scripts
    my $vault_dir = "$vhost_path/vault_scripts/$prefix";
    my $id = makeDumpFile("$dumpDir/sapp-scripts." . generateDumpId(), $vault_dir);
    if ($id) {
      $sapp_installed->{'ADDCHILD'}->(makeXmlNode('sapp-scripts', $id));
    }

	# sapp-installdir
	my $sapp_installdir = XmlNode('sapp-installdir',
				'children' => [XmlNode('sapp-prefix', 'content' => $prefix)]);
	if ($sapp->isSsl()) {
		$sapp_installdir->{'ADDCHILD'}->(XmlNode('sapp-ssl'));
	}
	$sapp_installed->{'ADDCHILD'}->($sapp_installdir);

	# cgi-bin
    my $cgibin_dir = "$vhost_path/cgi-bin/" . $prefix;
    $id = makeDumpFile("$dumpDir/sapp-cgi-bin." . generateDumpId(), $cgibin_dir, $sapp->{cgi_bin},
                      undef, undef, undef, 1);
    if ($id) {
      $sapp_installed->{'ADDCHILD'}->(makeXmlNode('sapp-cgi-bin', $id));
    }

	# htdocs
	my $htdocs_dir = $vhost_path;
	if ($sapp->isSsl()) {
		$htdocs_dir .= '/httpsdocs/';
	} else {
		$htdocs_dir .= '/httpdocs/';
	}
	$htdocs_dir .= $prefix;

    $id = makeDumpFile("$dumpDir/sapp-htdocs." . generateDumpId(), $htdocs_dir, $sapp->{htdocs},
                      undef, undef, undef, 1);
    if ($id) {
      $sapp_installed->{'ADDCHILD'}->(makeXmlNode('sapp-htdocs', $id));
	}

	return $sapp_installed;
}

sub makeDomainLogrotationNode {
  my ($parent, $domainId) = @_;

  if (PleskVersion::atLeast(5, 0, 0)) {
	my $sql = "SELECT val FROM dom_param WHERE param = 'logrotation_id' "
	  . "AND dom_id=$domainId";
	if ($wrapDbh->{'EXECUTE'}->($sql) and my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  my $logrotation_id = $ptrRow->[0];
	  $wrapDbh->{'FINISH'}->();
	  makeLogrotationNode($parent, $logrotation_id);
	}
	$wrapDbh->{'FINISH'}->();
  }
}

sub makeLogrotationNode {
  my ($parent, $logrotation_id) = @_;

  my $sql = "SELECT * FROM log_rotation WHERE id=$logrotation_id";

  if($wrapDbh->{'EXECUTE'}->($sql) and my $hashPtr = $wrapDbh->{'FETCHHASH'}->()) {
	my %logRotation = %{$hashPtr};

	if ($logRotation{'id'}) {
	  my $item = &makeXmlNode('logrotation');
	  $item->{'ATTRIBUTE'}->('max-number-of-logfiles',
							 $logRotation{'max_number_of_logfiles'});
	  $item->{'ATTRIBUTE'}->('compress', $logRotation{'compress_enable'});
	  unless (defined $logRotation{'email'}) {
		$logRotation{'email'} = '';
	  }
	  $item->{'ATTRIBUTE'}->('email', $logRotation{'email'});
	  $item->{'ATTRIBUTE'}->('enabled', $logRotation{'turned_on'});

	  my $rottype;
	  if($logRotation{'period_type'} eq 'by_time') {
		$rottype = &makeXmlNode('logrotation-period');
		$rottype->{'ATTRIBUTE'}->('period', $logRotation{'period'});
	  } else {
		if (PleskVersion::atLeast(7, 0, 0)) {
		  $rottype = &makeXmlNode('logrotation-maxsize', $logRotation{'period'});
		} else {
		  $rottype = &makeXmlNode('logrotation-maxsize', $logRotation{'period'}*1024);
		}
	  }
	  $item->{'ADDCHILD'}->($rottype);
	  $parent->{'ADDCHILD'}->($item);
	}
  }
  $wrapDbh->{'FINISH'}->();
}

#-----------------------------------------------------------------
#	$ptrSubDomain - pointer to hash with subdomain's info (row of table 'subdomains')
#	$domainName - name of domain
#	$ptrSiteApplications - pointer to list of domain's site applications
#-----------------------------------------------------------------
sub dumpSubDomain( $ $ $ )
{
	my ($ptrSubDomain, $domainName, $ptrSiteApplications) = @_;
	unless (ref($ptrSubDomain) =~ /HASH/) {
		printToError("Error: dumpSubDomain: bad argumets");
		return undef;
	}

	my $root = makeXmlNode('subdomain');

	my $subDomainName;
	if (PleskVersion::atLeast(7, 1, 0)) {
		$subDomainName = $ptrSubDomain->{'displayName'};
	} else {
		$subDomainName = $ptrSubDomain->{'name'};
	}
	$root->{'ATTRIBUTE'}->('name', $subDomainName);

	# Subdomains may be SSL-enabled since 8.0
	if (PleskVersion::atLeast(8, 0, 0)) {
		if ('true' eq $ptrSubDomain->{'ssl'}) {
			$root->{'ATTRIBUTE'}->('https', 'true');
		}
		if ('true' eq $ptrSubDomain->{'same_ssl'}) {
			$root->{'ATTRIBUTE'}->('shared-content', 'true');
		}
	}

	#-----------------------------------------------------------------
	# Site applications content should not be included into dump
	# together with other subdomain's content.
	#-----------------------------------------------------------------
	my @SiteApplications = @{$ptrSiteApplications};
	my @exclude_htdocs_files;
	my @exclude_htdocs_ssl_files;
	my @exclude_cgibin_files;
	foreach my $sapp (@SiteApplications) {
		if (!$sapp->isSubdomainSiteapp($ptrSubDomain->{'id'})) {
			next;
		}
		if ($sapp->isSsl()) {
			push @exclude_htdocs_ssl_files, $sapp->getHtdocsFiles();
		} else {
			push @exclude_htdocs_files, $sapp->getHtdocsFiles();
		}
		push @exclude_cgibin_files, $sapp->getCgibinFiles();
	}

	# content
	my $subDomainRoot = AgentConfig::get("HTTPD_VHOSTS_D") . "/$domainName/subdomains/$subDomainName";
	if (-d $subDomainRoot) {

		my $fileName;

		# httpdocs
		if ($fileName = makeDumpFile("$dumpDir/$subDomainName.$domainName.htdocs", "$subDomainRoot/httpdocs", '.', \@exclude_htdocs_files)) {
			$root->{'ATTRIBUTE'}->('cid_docroot', $fileName);
		}
		# httpsdocs
		if ($fileName = makeDumpFile("$dumpDir/$subDomainName.$domainName.shtdocs", "$subDomainRoot/httpsdocs", '.', \@exclude_htdocs_ssl_files)) {
			$root->{'ATTRIBUTE'}->('cid_docroot_ssl', $fileName);
		}
		# cgi-bin
		if ($fileName = makeDumpFile("$dumpDir/$subDomainName.$domainName.cgi", "$subDomainRoot/cgi-bin", '.', \@exclude_cgibin_files)){
			$root->{'ATTRIBUTE'}->('cid_cgi', $fileName);
		}
		# conf
		if ($fileName = makeDumpFile("$dumpDir/$subDomainName.$domainName.conf", "$subDomainRoot/conf", '.')) {
			$root->{'ATTRIBUTE'}->('cid_conf', $fileName);
		}
	}

	# sysuser
	if ($ptrSubDomain->{'sys_user_type'} =~ /native/) {
		my $item = makePleskSysUserNode($ptrSubDomain->{'sys_user_id'});
		if (ref($item) =~ /HASH/) {
			$root->{'ADDCHILD'}->($item);
		}
	}

	# scripting
	my $item = makeXmlNode('scripting');
	while (my ($xmlName, $fieldName) = each(%subDomainScripting)) {
		if ($ptrSubDomain->{$fieldName} =~ /true/) {
			$item->{'ATTRIBUTE'}->($xmlName, 'true');
		}
	}
	$root->{'ADDCHILD'}->($item);

	# Dump of installed site applications
	foreach my $sapp (@SiteApplications) {
		if (!$sapp->isSubdomainSiteapp($ptrSubDomain->{'id'})) {
			next;
		}
		$root->{'ADDCHILD'}->(dumpSiteApplication($sapp, $subDomainRoot));
	}

	return $root;
}


#
# makeWebUserNode
#
#  argumenets:
#       $ptrWebUser - pointer to hash with web user's info (row of table 'web_users')
#       $domainName - name of domain
#
#  return:
#       $root = XML node
#
sub makeWebUserNode {
  my($ptrWebUser,$domainName)=@_;
  unless(ref($ptrWebUser)=~/HASH/){
    &printToError("Error: makeWebUserNode: bad argumets");
    return undef;
  }
  my($root,$home,$fileName,$userName,$item,$xmlName,$fieldName);
  $root = &makeXmlNode('webuser');


  if (PleskVersion::atLeast(5, 0, 0)) {
	$item = &makePleskSysUserNode($ptrWebUser->{'sys_user_id'});
	if(ref($item)=~/HASH/){
	  $root->{'ADDCHILD'}->($item);
	  $userName = $item->{'ATTRIBUTE'}->('name');
	  $root->{'ATTRIBUTE'}->('name',$userName);
    }
  } else {
	$userName = $ptrWebUser->{'username'};
	my @pwnam = getpwnam($userName);
	my $password = $pwnam[1];
	$item = &makeSyntheticSysUserNode($userName, $password, 'encrypted');
	$root->{'ADDCHILD'}->($item);
	$root->{'ATTRIBUTE'}->('name',$userName);
  }

  my $subDomainRoot=AgentConfig::get("HTTPD_VHOSTS_D")."/$domainName/web_users/$userName";

  if(-d $subDomainRoot){
    if($fileName = &makeDumpFile("$dumpDir/$domainName\_$userName",
        $subDomainRoot)){
      $root->{'ATTRIBUTE'}->('cid',$fileName);
    }
  }

  my $privateData = AgentConfig::get("HTTPD_VHOSTS_D")."/$domainName/private/$userName";
  if (-d $privateData) {
    if ($fileName = makeDumpFile("$dumpDir/$domainName\_$userName.private",
        $privateData)) {
      $root->{'ATTRIBUTE'}->('cid_private', $fileName);
    }
  }

#
# scripting
#
  $item = &makeXmlNode('scripting');
  while(($xmlName,$fieldName)=each(%webUserScripting)){
    if (defined $ptrWebUser->{$fieldName} and $ptrWebUser->{$fieldName}=~/true/) {
      $item->{'ATTRIBUTE'}->($xmlName,'true');
    }
  }
  $root->{'ADDCHILD'}->($item);
#
# end scripting
#

  return $root;
}

#
# makeProtDirNode - make protected directory node
#
# arguments:
#        $pdirId - ID
#        $pdirPath - path of protected directory
#        $pdirTitle - header text
#        $pdirNonSSL - whether is directory in http
#        $pdirSSL - or in https
#
# return:
#        $root - XML node
#
sub makeProtDirNode {
  my($pdirId, $pdirPath, $pdirTitle, $pdirNonSSL, $pdirSSL, $pdirCGI) = @_;

  my($root, $sql, $userNode, $item, $ptrRow);

  $root = makeXmlNode('pdir');
  # workaround of CLI inabliity to create '' directory.
  if ($pdirPath eq '') {
	$pdirPath = '/';
  }

  $root->{'ATTRIBUTE'}->('name', $pdirPath);
  if($pdirTitle) {
    $root->{'ATTRIBUTE'}->('title', $pdirTitle);
  }
  if ($pdirNonSSL) {
    $root->{'ATTRIBUTE'}->('nonssl', $pdirNonSSL);
  }
  if ($pdirSSL) {
    $root->{'ATTRIBUTE'}->('ssl', $pdirSSL);
  }
  if ($pdirCGI) {
	$root->{'ATTRIBUTE'}->('cgi', $pdirCGI);
  }

  if (PleskVersion::atLeast(7, 1, 0)) {
    $sql = "SELECT p.login, a.password, a.type FROM pd_users p ".
      " LEFT JOIN accounts a ON p.account_id = a.id ".
	  " WHERE p.pd_id=$pdirId ORDER BY p.id";
  } else {
    $sql = "SELECT p.login, p.passwd, 'plain' FROM pd_users p ".
      " WHERE p.pd_id=$pdirId ORDER BY p.id"
  }

  if ($wrapDbh->{'EXECUTE'}->($sql)) {
    while ($ptrRow=$wrapDbh->{'FETCHROW'}->()) {
      $userNode = makeXmlNode('pduser');
      $root->{'ADDCHILD'}->($userNode);
      $userNode->{'ATTRIBUTE'}->('name', $ptrRow->[0]);
      $item = makePasswordNode($ptrRow->[1], pwdType($ptrRow->[2]));
      $userNode->{'ADDCHILD'}->($item);
    }
  }
  $wrapDbh->{'FINISH'}->();
  return $root;
}

sub makeAnonftpPermissionNode {
  my ($parent, $name) = @_;
  my $node = makeXmlNode('anonftp-permission');
  $node->{'ATTRIBUTE'}->('name', $name);
  $parent->{'ADDCHILD'}->($node);
}

sub makeAnonftpLimitNode {
  my ($parent, $name, $value) = @_;
  if($value != 0) {
	my $node = makeXmlNode('anonftp-limit', $value);
	$node->{'ATTRIBUTE'}->('name', $name);
	$parent->{'ADDCHILD'}->($node);
  }
}

sub addAnonFtp {
  my($root,$domainId,$domainName)=@_;
  unless(ref($root)=~/HASH/){
    &printToError("Error: addAnonFtp: bad argumets");
    return undef;
  }

  my($anonRoot,$ptrHash,$count,$fileName,$domainRoot,$sql);
  $count=0;
  $sql="SELECT * FROM anon_ftp WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    while($ptrHash=$wrapDbh->{'FETCHHASH'}->()){

      $anonRoot = &makeXmlNode('anonftp');
      $root->{'ADDCHILD'}->($anonRoot);
      if($ptrHash->{'status'}=~/true/){
		$anonRoot->{'ATTRIBUTE'}->('pub','true');
      }
      if($ptrHash->{'incoming'}=~/true/){
		$anonRoot->{'ATTRIBUTE'}->('incoming','true');
      }
	  if (defined($ptrHash->{'max_conn'})) {
		makeAnonftpLimitNode($anonRoot,'max-connections', $ptrHash->{'max_conn'});
	  }
	  if (defined($ptrHash->{'bandwidth'})) {
		makeAnonftpLimitNode($anonRoot, 'bandwidth', $ptrHash->{'bandwidth'});
	  }
	  if (defined($ptrHash->{'quota'})) {
		makeAnonftpLimitNode($anonRoot, 'incoming-disk-quota', $ptrHash->{'quota'});
	  }
	  if (defined($ptrHash->{'display_login'})) {
		$anonRoot->{'ATTRIBUTE'}->('display-login', $ptrHash->{'display_login'});
	  }
	  if ($ptrHash->{'incoming_readable'} =~ /true/) {
		makeAnonftpPermissionNode($anonRoot, 'incoming-download');
	  }
	  if ($ptrHash->{'incoming_subdirs'} =~ /true/) {
		makeAnonftpPermissionNode($anonRoot, 'incoming-mkdir');
	  }
	  if (defined($ptrHash->{'login_text'})) {
		my $loginMessageNode = makeXmlNode('login-message', $ptrHash->{'login_text'});
		$anonRoot->{'ADDCHILD'}->($loginMessageNode);
	  }
      $domainRoot=AgentConfig::get("HTTPD_VHOSTS_D").'/'.$domainName;
      if(-d $domainRoot){

		if($fileName = &makeDumpFile("$dumpDir/$domainName.anonftp.pub",
									 "$domainRoot/anon_ftp/pub")){
		  $anonRoot->{'ATTRIBUTE'}->('cid',$fileName);
		}
		if($fileName = &makeDumpFile("$dumpDir/$domainName.anonftp.incoming",
									 "$domainRoot/anon_ftp/incoming")){
		  $anonRoot->{'ATTRIBUTE'}->('cid_incoming',$fileName);
		}
      }
      $count++;
    }
  }
  $wrapDbh->{'FINISH'}->();
  return $count;
}

sub getSysUserInfo {
  my ($sysUserId) = @_;

  my $sql = "SELECT * FROM sys_users WHERE id=$sysUserId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    printToError("Error: makeSysUser: sys user #$sysUserId  is not found");
    return undef;
  }
  my %sysuser = %{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();

  return %sysuser;
}

sub makePleskSysUserNode {
  my %sysuser = getSysUserInfo(@_);

  Logging::trace("Making system user node: $sysuser{'login'}");

  my $root = &makeXmlNode('sysuser');
#
# attributes
#
  $root->{'ATTRIBUTE'}->('name',lc($sysuser{'login'}));
  $root->{'ATTRIBUTE'}->('shell',$sysuser{'shell'}) if $sysuser{'shell'};

  my $quota;
  if($sysuser{'quota'}){
	if (PleskVersion::atLeast(7, 0, 0)) {
	  $quota = $sysuser{'quota'};
	} else {
	  if ($sysuser{'quota'} ne "-1") {
		$quota = $sysuser{'quota'}*1024*1024;
	  }
	}
	$root->{'ATTRIBUTE'}->('quota',$quota) if $quota;
  }

#
# end attributes
#

#
# password node
#
  if (PleskVersion::atLeast(6, 0, 0)) {
	$root->{'ADDCHILD'}->(makeAccountPasswordNode($sysuser{'account_id'}));
  } else {
	$root->{'ADDCHILD'}->(makePasswordNode($sysuser{'passwd'}, 'plain'));
  }
#
# end password node
#

#
# crontab node
#
  my $crontabmng = AgentConfig::get("PRODUCT_ROOT_D") . "/admin/sbin/crontabmng";
  if (-x $crontabmng) {
	open (CRONTAB, "$crontabmng get $sysuser{'login'} |");
	my $crontab = "";
	while (<CRONTAB>) {
	  last if $_ eq "0\n";
	  $crontab .= $_;
	}
	close (CRONTAB);
	if ($crontab ne "\n") {
	  $root->{'ADDCHILD'}->(makeXmlNode('cron', $crontab));
	}
  }
#
# end crontab node
#
  return $root;

}

sub makeSyntheticSysUserNode {
  my ($name, $password, $passtype) = @_;

  my $root = &makeXmlNode('sysuser');
  $root->{'ATTRIBUTE'}->('name', lc($name));
  my $passwdNode = &makePasswordNode($password, $passtype);
  $root->{'ADDCHILD'}->($passwdNode);
  return $root;
}

sub makeFhostingNode {
  my ($ptrDomain)=@_;
  unless(ref($ptrDomain)=~/HASH/){
    &printToError("Error: makeFhostingNode: bag arguments");
    return undef;
  }
  my($sql,$domainId,$root,$forward);
  $domainId=$ptrDomain->{'id'};
  $sql="SELECT redirect FROM forwarding WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    ($forward)=@{$wrapDbh->{'FETCHROW'}->()};
    $root=&makeXmlNode('fhosting',$forward);
  }else{
    &printToError("Error: makeFhostingNode: forward for domain '".$ptrDomain->{'name'}."' is not found");
  }
  $wrapDbh->{'FINISH'}->();

  return $root;
}

sub makeShostingNode {
  my ($ptrDomain)=@_;
  unless(ref($ptrDomain)=~/HASH/){
    &printToError("Error: makeShostingNode: bag arguments");
    return undef;
  }

  my($sql,$domainId,$root,$forward);
  $domainId=$ptrDomain->{'id'};
  $sql="SELECT redirect FROM forwarding WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    ($forward)=@{$wrapDbh->{'FETCHROW'}->()};
    $root=&makeXmlNode('shosting',$forward);
  }else{
    &printToError("Error: makeShostingNode: forward for domain '".$ptrDomain->{'name'}."' is not found");
  }
  $wrapDbh->{'FINISH'}->();

  return $root;
}

sub getClientIps {
  my ($clientId) = @_;

  if (PleskVersion::atLeast(6, 0, 0)) {
  } elsif (PleskVersion::atLeast(5, 0, 0)) {
    my $sql = "SELECT d.name, d.id, h.ip_address, a.status, d.cl_id FROM domains d "
      . "LEFT JOIN hosting h ON h.dom_id = d.id "
      . "LEFT JOIN anon_ftp a ON a.dom_id = d.id"
      . "WHERE cl_id = $clientId";
  } else {
  }
}

#
# Returns password node for given account ID
#
sub makeAccountPasswordNode
{
	my $account_id = shift;
	my $node;

	if ($wrapDbh->{'EXECUTE'}->("SELECT password, type FROM accounts WHERE id='$account_id'")) {
		my ($passwd, $type) = @{$wrapDbh->{'FETCHROW'}->()};
		$node = &makePasswordNode($passwd, pwdType($type));
    } else {
		# generates a stub node
		$node = &makePasswordNode();
	}
	$wrapDbh->{'FINISH'}->();

	return $node;
}

sub printAgentStatus {
  my $root = makeXmlNode('agent-status');

  unless (PleskVersion::get() && (defined $bin{'iconv'}) && PleskVersion::isSupported()) {
    my $item = XmlNode('wrong-platform');
	if (!PleskVersion::isSupported()) {
      $item->{'CONTENT'}->("Migration from Plesk versions above " . PleskMaxVersion::get() . " is not supported");
	} elsif (not defined $bin{'iconv'}) {
      $item->{'CONTENT'}->("no iconv found on the source host");
	}

    $root->{'ADDCHILD'}->($item);
  }

  printXml($root, *STDOUT);
}

# Converts 'crypt' -> 'encrypted'
# It's a workaround for psa.acounts.type
# bug 47030
sub pwdType
{
	my ($type) = @_;
	$type =~ s/^crypt$/encrypted/;
	return $type;
}

sub makeSpamassassinNode {
	my ($root,$mailName)=@_ ;

  my  $sql = "SELECT count(*) FROM sa_conf WHERE mailname='$mailName'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about spam filter for $mailName ");

  return undef;

  }
  my @row = @{$wrapDbh->{'FETCHROW'}->()};
  $wrapDbh->{'FINISH'}->();
  if ($row[0]==0) {
      return;
  }

  $sql = "SELECT * FROM sa_conf WHERE mailname='$mailName'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    &printToError("Error during getting information about spam filter for $mailName ");

  return undef;
  }
  my %spam = %{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();

	my $item;
    if ($spam{'flt_enabled'} eq 'both'){
		$item=&makeXmlNode('spamassassin');
		$item->{'ATTRIBUTE'}->('status','on');
		$item->{'ATTRIBUTE'}->('server-conf','true');
	} elsif ($spam{'flt_enabled'} eq 'user'){
		$item=&makeXmlNode('spamassassin');
		$item->{'ATTRIBUTE'}->('status','on');
		$item->{'ATTRIBUTE'}->('server-conf','false');
	} else {
		#skip creation
		return ;
	}
	$item->{'ATTRIBUTE'}->('hits',$spam{'hits_required'});
		if($spam{'spam_action'} eq 'save'){
			$item->{'ATTRIBUTE'}->('action','mark');
		}else{
			$item->{'ATTRIBUTE'}->('action','del');
		}

		if($spam{'rw_subject_tag'}){
			$item->{'ATTRIBUTE'}->('modify-subj','true');
			$item->{'ATTRIBUTE'}->('subj-text',$spam{'rw_subject_tag'});
		}else{
			$item->{'ATTRIBUTE'}->('modify-subj','false');
		}

	&makeBlackListNode($item,$spam{'mailname'});
	&makeWhiteListNode($item,$spam{'mailname'});

	$root->{'ADDCHILD'}->($item);
	return $root;

}

sub makeWhiteListNode {
	my ($root,$mailname) =@_;
	my ($sql,@patterns,$pattern,$item,$ptrRow);
   $sql = "SELECT count(pattern) FROM sa_list WHERE mailname='$mailname' AND color='white'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about white list for $mailname ");
  return undef;
  }
	$ptrRow = $wrapDbh->{'FETCHROW'}->();
	if ($ptrRow->[0]==0){
		$wrapDbh->{'FINISH'}->();
        return;
	}

$wrapDbh->{'FINISH'}->();

  $sql = "SELECT pattern FROM sa_list WHERE mailname='$mailname' AND color='white'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about white list patterns for $mailname ");
  return undef;
  }
  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
		push @patterns,$ptrRow->[0];
	  }
	  foreach $pattern (@patterns){
	$item = makeXmlNode('whitelist-member',$pattern);
	$root->{'ADDCHILD'}->($item);
	}
	$wrapDbh->{'FINISH'}->();
	return $root;
}

sub makeBlackListNode {
	my ($root,$mailname)=@_ ;
	my ($sql,@patterns,$pattern,$item,$ptrRow);
   $sql = "SELECT count(pattern) FROM sa_list WHERE mailname='$mailname' AND color='black'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about black list for $mailname ");
  return undef;
  }
$ptrRow = $wrapDbh->{'FETCHROW'}->();
   if ($ptrRow->[0]==0){
	$wrapDbh->{'FINISH'}->();
   	return;
   }

$wrapDbh->{'FINISH'}->();

   $sql = "SELECT pattern FROM sa_list WHERE mailname='$mailname' AND color='black'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about black list patterns for $mailname ");
  return undef;
  }
  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
		push @patterns,$ptrRow->[0];
	  }
	  foreach $pattern (@patterns){
	$item = makeXmlNode('blacklist-member',$pattern);
	$root->{'ADDCHILD'}->($item);
	}
	$wrapDbh->{'FINISH'}->();
	return $root;

}

sub makeDrwebNode {
  my ($root, $mailname)=@_;
  my $drwebFile = "/etc/drweb/users.conf";
  my $item;
  unless (open(DRWEBFILE, "<$drwebFile")) {
	printToLog("Error: unable to open $drwebFile");
	return;
  }
  while (<DRWEBFILE>) {
	chomp;
	next if /^#/; #skip comments
	next if /^\[/;
	next if /^\s*$/;
	my @parts = split /\s+/, $_, 4;
	if ($parts[3] =~ /$mailname/) {
	  $item = &makeXmlNode('drweb');
	  if ($parts[1] =~ /to/) {
		$item->{'ATTRIBUTE'}->('state', 'in');
	  } elsif ($parts[1] =~ /from/) {
		$item->{'ATTRIBUTE'}->('state', 'out');
	  } else {
		$item->{'ATTRIBUTE'}->('state', 'inout');
	  }
	}
  }
  close(DRWEBFILE);
  $root->{'ADDCHILD'}->($item);
  return $root;
}

sub makeWebAppNode {
  my ($root,$webAppName,$webAppStatus,$dom_name)=@_;

  # webapp is useless without cid
  return if exists $ptrArgs->{'configuration-only'};

  $webAppStatus ||= 0;

  my $item = XmlNode('webapp',
					 'attributes' => {'name' => $webAppName},
					 'children' => [Status::make($webAppStatus)]);

  my $warDir = AgentConfig::get("CATALINA_HOME") . "/psa-wars/$dom_name";
  my $warFile = "$webAppName.war";
  my $dumpFile = makeDumpFile("$dumpDir/$dom_name.$webAppName.webapp",$warDir,$warFile);

  if ($dumpFile) {
  	$item->{'ATTRIBUTE'}->('cid',$dumpFile);
  }

  $root->{'ADDCHILD'}->($item);
}

sub parseCustomButtonOptions {
  my ($node, $options) = @_;

  #CUSTOM_BUTTON_PUBLIC
  $node->{'ATTRIBUTE'}->('visible-to-sublogins',
						 $options & 128 ? 'true' : 'false');
  #CUSTOM_BUTTON_INTERNAL
  $node->{'ATTRIBUTE'}->('open-in-same-frame',
						 $options & 256 ? 'true' : 'false');

  my %options =
	(
	 "domain-id" => 1, #CUSTOM_BUTTON_DOM_ID
	 "domain-name" => 32, #CUSTOM_BUTTON_DOM_NAME
	 "ftp-login" => 512, #CUSTOM_BUTTON_FTP_USER
	 "ftp-password" => 1024, #CUSTOM_BUTTON_FTP_PASS
	 "client-id" => 2, #CUSTOM_BUTTON_CL_ID
	 "client-company-name" => 4, #CUSTOM_BUTTON_CNAME
	 "client-contact-name" => 8, #CUSTOM_BUTTON_PNAME
	 "client-email" => 16, #CUSTOM_BUTTON_EMAIL
	);

  while (my ($optname, $optmask) = each %options) {
	if ($options & $optmask) {
	  my $optnode = makeXmlNode('url-option');
	  $optnode->{'ATTRIBUTE'}->('name', $optname);
	  $node->{'ADDCHILD'}->($optnode);
	}
  }
}

sub getCustomButtonById71( $ )
{
	my ($id) = @_;

	my $sql = "SELECT sort_key, place, text, url, conhelp, options, file FROM custom_buttons WHERE id=$id";
	if (!$wrapDbh->{'EXECUTE'}->($sql)) {
		$wrapDbh->{'FINISH'}->();
		return;
	}

	my $ptrHash = $wrapDbh->{'FETCHHASH'}->();
	$wrapDbh->{'FINISH'}->();

	my $node = makeXmlNode('custom-button');

	my %attributes = ( 'url' => 'url',
					   'text' => 'text',
					   'sort_key' => 'sort-priority',
					   'place' => 'interface-place',
					   'conhelp' => 'conhelp' );

	while (my ($dbfield, $nodefield) = each %attributes) {
		$node->{'ATTRIBUTE'}->($nodefield, $ptrHash->{$dbfield});
	}

	my $customButtonsDir = AgentConfig::get("PRODUCT_ROOT_D") . '/admin/htdocs/images/custom_buttons';

	my $icon = $ptrHash->{'file'};
        if (-e $customButtonsDir."/".$icon) {
	  if ($icon && (my $cid_icon = makeDumpFile("$dumpDir/$icon.$id", $customButtonsDir, $icon))) {
	    $node->{'ATTRIBUTE'}->('cid-icon', $cid_icon);
	  }
        }
	parseCustomButtonOptions($node, $ptrHash->{'options'});

	return $node;
}

sub getCustomButtonsByOwner71
{
	my ($owner_type, $owner_id) = @_;

	my $level = $custom_button_owner_types{$owner_type};
	my $sql = "SELECT id FROM custom_buttons WHERE level='$level' AND level_id='$owner_id'";
	if (!$wrapDbh->{'EXECUTE'}->($sql)) {
		$wrapDbh->{'FINISH'}->();
		return;
	}

	my @ids;
	while (my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		my $id = $ptrRow->[0];
		if (grep /^$id$/, @skip_custom_buttons) {
			next;
		}
		push @ids, $id;
	}
	$wrapDbh->{'FINISH'}->();

	my @nodes;
	foreach my $id (@ids) {
		push @nodes, getCustomButtonById71($id);
	}
	return @nodes;
}

sub getCustomButtonsByOwner65 {
  my ($owner_type, $owner_id) = @_;

  return if ($owner_type ne "client");

  my $sql = "SELECT text, url, conhelp, options FROM custom_buttons WHERE " .
	"cl_id=$owner_id";

  my @nodes;

  if ($wrapDbh->{'EXECUTE'}->($sql)) {

	my $ptrHash;
	while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
	  my $node = makeXmlNode('custom-button');
	  $node->{'ATTRIBUTE'}->('sort-priority', '100');
	  $node->{'ATTRIBUTE'}->('interface-place', 'domain');

	  my %attributes =
		(
		 'url' => 'url',
		 'text' => 'text',
		 'conhelp' => 'conhelp'
		);

	  while (my ($dbfield, $nodefield) = each %attributes) {
		$node->{'ATTRIBUTE'}->($nodefield, $ptrHash->{$dbfield});
	  }

	  ## 128 == CUSTOM_BUTTON_PUBLIC
	  parseCustomButtonOptions($node, $ptrHash->{'options'} | 128);

	  push @nodes, $node;
	}
  }
  $wrapDbh->{'FINISH'}->();
  return @nodes;
}

sub getCustomButtonsByOwner {
  return getCustomButtonsByOwner71(@_) if (PleskVersion::atLeast(7, 1, 0));
  return getCustomButtonsByOwner65(@_) if (PleskVersion::atLeast(6, 5, 0));
}

sub getPostgresVersion {
  my $psql = $bin{'psql'};
  my @out = `$psql --version | awk '{print \$3}'`;
  chomp $out[0];
  if ($out[0]=~/(\d+\.\d+\.\d+).*/){
    return $1;
  }
}

sub makeServerNode {
  my ($parent) = @_;
  my $serverNode = XmlNode('server');

  # Dump database servers
  if(PleskVersion::atLeast(8,0,0)){
    my $sql = "SELECT param, val FROM misc WHERE param like 'default_server_%'";
    my %def;

    $wrapDbh->{'EXECUTE'}->($sql);
    while (my $ptrHash = $wrapDbh->{'FETCHHASH'}->()){
       $def{$ptrHash->{'param'}} = $ptrHash->{'val'};
    }

    $sql = "SELECT id, host, port, type, admin_login, admin_password FROM DatabaseServers";
    $wrapDbh->{'EXECUTE'}->($sql);

    while (my $ptrHash = $wrapDbh->{'FETCHHASH'}->()){
      my $dbServerNode = XmlNode('db-server');
      my $param = 'default_server_'.$ptrHash->{'type'};

      if (exists $def{$param} && $def{$param} == $ptrHash->{'id'}) {
        $dbServerNode->{'ATTRIBUTE'}->('default', 'true');
      }
      $dbServerNode->{'ATTRIBUTE'}->('type',$ptrHash->{'type'});

      $dbServerNode->{'ADDCHILD'}->(XmlNode('host', 'content' => "$ptrHash->{'host'}"));
      $dbServerNode->{'ADDCHILD'}->(XmlNode('port', 'content' => "$ptrHash->{'port'}"));

      if ($ptrHash->{'admin_login'} ne '') {
        my $adminNode = XmlNode('admin',"attributes" => {"name" => "$ptrHash->{'admin_login'}"});
        if ($ptrHash->{'type'} eq 'mysql' && $ptrHash->{'host'} eq 'localhost') {
          $ptrHash->{'admin_password'} = AgentConfig::get('password');
        }
        my $passwordNode = makePasswordNode($ptrHash->{'admin_password'},'plain'); # password type for dbservers is always plain
        $adminNode->{'ADDCHILD'}->($passwordNode);

        $dbServerNode->{'ADDCHILD'}->($adminNode);
      }

      $serverNode->{'ADDCHILD'}->($dbServerNode);
    }
    $wrapDbh->{'FINISH'}->();
  }

  if (exists $ptrArgs->{'no-content'} or exists $ptrArgs->{'configuration-only'}) {
    $parent->{'ADDCHILD'}->($serverNode);
    return;
  }

  # Dump key info
  my $sql = "SELECT * FROM key_history WHERE filename IS NOT NULL";
  $wrapDbh->{'EXECUTE'}->($sql);

  while (my $ptrHash = $wrapDbh->{'FETCHHASH'}->()){
    next if ($ptrHash->{'filename'}=~/\~$/);
    my @keyPath = split(/\//,$ptrHash->{'filename'});
    my ($keyName,$keyDir,$additional);
    if ($keyPath[3] eq 'key.d') {
      $keyName = $keyPath[4];
      $keyDir = "/etc/psa/key.d/";
      $additional = "true";
    } else {
      $keyName = $keyPath[3];
      $keyDir = "/etc/psa/";
      $additional = "false";
    }

    my $keyNode;
    
    if (-e $keyDir."/".$keyName) {
      my $cid = makeDumpFile("$dumpDir/$ptrHash->{'plesk_key_id'}.key", $keyDir, $keyName);

      $keyNode = XmlNode('key', 'attributes' => {'additional' => $additional});
   
      $keyNode->{'ATTRIBUTE'}->('cid', $cid) if $cid;

      $serverNode->{'ADDCHILD'}->($keyNode);
    }
  }

  $wrapDbh->{'FINISH'}->();
  $parent->{'ADDCHILD'}->($serverNode);
}

sub printHelp
{
  print <<HELP

Usage:
  $0 <options>

Options:
  -s |--get-status           get status of the agent
  -dc|--dump-accounts=<list> comma-separated list of clients to dump
  -dd|--dump-domains=<list>  comma-separated list of domains to dump
  -da|--dump-all             dump all objects on hosts
      --server               dump server settings

  -lc|--get-content-list     get list of content files
  -nc|--no-content           make a shallow structure dump
  -co|--configuration-only   dump only structure
  -om|--only-mail            dump only mail content
  -nz|--no-compress          do not compress content files

  -o|--output=<file>        'tar' the dump files to the single file
                             (splitting on 2040M boundary and appending
                             .001, .002, ... suffixes)

  -d|--description=<text>    Add description to the dump

  -h |--help                 this help

HELP
;
}
